from src import generate_metafile
import src.utils as utils
import copy
import os


def edit_file(project_id, path, mode, mandatory_only, size=80):

    if mode == 'metadata':
        key_yaml = utils.read_in_yaml(os.path.join(os.path.dirname(
            os.path.abspath(__file__)), '..', 'keys.yaml'))
    else:
        key_yaml = utils.read_in_yaml(os.path.join(os.path.dirname(
            os.path.abspath(__file__)), '..', 'mamplan_keys.yaml'))

    file = utils.read_in_yaml(path)

    options = [key for key in key_yaml]

    print(f'Choose the parts you want to edit (1,...,{len(options)}) divided '
          f'by comma.\n')

    generate_metafile.print_option_list(options, False)
    edit_keys = generate_metafile.parse_input_list(options, True)

    for key in edit_keys:

        if key in file:
            file[key] = edit_item(project_id, key, file[key], key_yaml[key], key_yaml,
                                  file, mandatory_only, mode)

        else:

            file[key] = generate_metafile.get_redo_value(project_id,
                key_yaml[key], key, True, mandatory_only, file, True, False,
                True, mode, key_yaml)

        while True:

            print(generate_metafile.get_summary(file[key], size=size))

            correct = generate_metafile.parse_list_choose_one(
                ['True ', 'False '], f'\nIs the input correct? You can redo '
                                     f'it by selecting \'False\'')

            if correct:
                break

            else:
                file[key] = edit_item(project_id, key, file[key], key_yaml[key], key_yaml,
                                      file, mandatory_only, mode, size=size)

    utils.save_as_yaml(file, path)
    print(f'Changes were saved to {path}')


# TODO: split into multiple functions
def edit_item(project_id, key_name, item, key_yaml, full_yaml, result_dict, mandatory_mode,
              mode, size=80, not_editable=None):
    """
    This function is used to edit a dictionary. Therefor it requests
    information from the user via a dialog
    :param key_name: The name of the key whose values should be edited
    :param item: the value to be edited
    :param key_yaml: a part of the underlying structure containing the key
                     whose values should be edited
    :param full_yaml: the complete general structure
    :param result_dict: the filled dictionary
    :param mandatory_mode: a bool that states if the mandatory mode is active
    :param mode: the mode the program runs in ('metadata' or 'mamplan')
    :param size: the window size (default 80)
    :param not_editable: a list of keys that is not editable
    :return: item: the edited value
    """

    # TODO: handle special cases

    if not_editable is None:
        not_editable = []

    # test if item to edit is a list
    if isinstance(item, list):

        # test that the list elements are not of type dict or if they are that
        # they are treated as 'merge' special case
        if not isinstance(key_yaml['value'], dict):

            # call 'get_redo_value' function to repeat the input
            item = generate_metafile.get_redo_value(project_id,
                key_yaml, key_name, not key_yaml['mandatory'], mandatory_mode,
                result_dict, True, False, True, mode, key_yaml)

        # list elements contain dictionary
        else:

            # initialize a list to hold list elements as options for the user
            # to choose to edit and set the first element to an option to
            # remove list elements
            all_options = ['remove element from list']

            # iterate over the list
            for i in range(len(item)):

                # TODO: enhance display of dictionary

                # if the item is a dictionary, convert it to a string
                # add the item to the option list
                if isinstance(item[i], dict):
                    str_dict = '\n'.join(f'{x}: {item[i][x]}' for x in item[i])
                    all_options.append(f'edit: {str_dict}')
                else:
                    all_options.append(f'edit: {item[i]}')

            # add the option to add a new element to the list
            all_options.append('add element to list')

            # request user input
            print(f'Please choose the how you want to edit the list choosing '
                  f'from the following options (1-{len(item)}) divided by '
                  f'comma.')

            # print the list elements as options and parse the user input
            generate_metafile.print_option_list(all_options, False)
            chosen_options = generate_metafile.parse_input_list(all_options,
                                                                False)

            # remove the options that do not represent a distinct list element
            all_options.remove('remove element from list')
            all_options.remove('add element to list')

            # initialize a list to store list elements that should be removed
            remove_options = []

            # test if user chose to remove elements
            if 'remove element from list' in chosen_options:

                # request user to state elements to delete
                print(f'Please choose the list elements you want to remove'
                      f' (1-{len(item)}) divided by comma.')

                # print a list of removable elements and parse user input
                generate_metafile.print_option_list(all_options, False)
                remove_options = generate_metafile.parse_input_list(
                    all_options, False)

            # initialize a dictionary to save editing information for all list
            # elements
            edit_options = {}

            # iterate over all list elements
            for i in range(len(all_options)):

                # define an action for the element depending on the users input
                # possible values are:
                # 'remove': if the element should be removed from the list
                # 'edit': if the element should be edited
                # None: if the element should stay as it is
                if all_options[i] in remove_options:
                    action = 'remove'
                elif all_options[i] in chosen_options:
                    action = 'edit'
                else:
                    action = None

                # add the list element and its according action to the dict
                edit_options[all_options[i]] = {'element': item[i],
                                                'action': action}

            # initialize a new list to save all elements to which do not get
            # removed (edited as well as not edited)
            new_list = []

            # iterate over all list elements
            for key in edit_options:

                # test if the element should be edited
                if edit_options[key]['action'] == 'edit':

                    # TODO: enhance line breaks (if display name > size)

                    # print header for current element
                    display_name = key.replace('\n', ' | ')
                    print(f'\n'
                          f'{"".center(size, "-")}\n'
                          f'{f"{display_name}".center(size, " ")}\n'
                          f'{"".center(size, "-")}\n')

                    # call this function to overwrite the list element with its
                    # edited version
                    edit_options[key]['element'] = edit_item(project_id,
                        key_name, edit_options[key]['element'], key_yaml,
                        full_yaml, result_dict, mandatory_mode, mode)

                # test if the list element should NOT be removed
                if edit_options[key]['action'] != 'remove':

                    # save the list element to the new list
                    new_list.append(edit_options[key]['element'])

            # overwrite the list with the edited list
            item = new_list

            # test if the user chose to add new list elements
            if 'add element to list' in chosen_options:

                # set the display_name and print it as header for the new
                # element
                display_name = key_yaml['display_name']
                print(f'\n'
                      f'{"".center(size, "-")}\n'
                      f'{f"New {display_name}".center(size, " ")}\n'
                      f'{"".center(size, "-")}\n')

                # get input for new element
                item += generate_metafile.get_redo_value(project_id,
                    key_yaml, key_name, not key_yaml['mandatory'],
                    mandatory_mode, result_dict, True, False, True, mode, key_yaml)

    # item to edit is a dictionary
    elif isinstance(item, dict):

        # request input from the user
        print(f'Please choose the keys (1-{len(item.keys())}) you want to edit'
              f' divided by comma.')

        # TODO: remove not_editable

        # create a list of options from the dictionary keys (only add them if
        # they are supposed to be edited)
        edit_index = [key for key in key_yaml['value'] if key not in
                      not_editable]

        # add option 'all' to redo the complete dictionary
        edit_index.insert(0, 'all')

        # print options for the user and parse the given input
        generate_metafile.print_option_list(edit_index, False)
        edit_index = generate_metafile.parse_input_list(edit_index, False)

        # test if 'all' was selected
        if 'all' in edit_index:

            # redo input for the whole dictionary
            new_item = generate_metafile.get_redo_value(project_id,
                key_yaml, key_name, not key_yaml['mandatory'], mandatory_mode,
                result_dict, True, False, False, mode, key_yaml)

            # input value is a list
            if isinstance(new_item, list):

                # overwrite the old value with the new one
                item = new_item

            # input value is a dictionary
            else:

                # combine and update the old value with the new one
                item = {**result_dict[key_name],
                        **new_item}

        # keys were selected but not 'all'
        else:

            # iterate over keys
            for key in edit_index:

                # the key 'organism' was selected (part experimental_setting)
                if key == 'organism':

                    # redo the whole dictionary because the latter parts depend
                    # on the organism
                    item = generate_metafile.get_redo_value(project_id,
                        key_yaml, 'experimental_setting',
                        not key_yaml['mandatory'], mandatory_mode, result_dict,
                        True, False, False, mode, key_yaml)[0]

                # the key 'experimental_factors was selected and not the key
                # 'organism' (part experimental_setting)
                elif key == 'experimental_factors' and 'organism' not in \
                        edit_index:

                    # copy the underlying structure (new_yaml) and remove the
                    # key 'organism'
                    new_yaml = copy.deepcopy(key_yaml)
                    new_yaml['value'].pop('organism')

                    # redo all except 'organism' since all keys below the key
                    # 'experimental_factors' depend on its values
                    item = generate_metafile.get_redo_value(project_id,
                        new_yaml, 'experimental_setting',
                        not key_yaml['mandatory'], mandatory_mode,
                        copy.deepcopy(item), True, False, False, mode, key_yaml)[0]

                # the key 'conditions was selected and not the keys 'organism'
                # or 'experimental_factors'
                elif key == 'conditions' and 'experimental_factors' not in \
                        edit_index and 'organism' not in edit_index:

                    # The following loop is needed so get the experimental
                    # factors into the structure they were directly after input
                    # before parsing into the yaml structure. This is needed
                    # for the conditions (combinations of all selected factors
                    # and values) to be created again.

                    # iterate over the experimental factors
                    for i in range(len(item['experimental_factors'])):

                        # extract the properties of the experimental factor
                        # from the underlying structure
                        fac_yaml = list(utils.find_keys(
                            full_yaml,
                            item['experimental_factors'][i]['factor']))

                        # test if the factor was found in the general structure
                        if len(fac_yaml) > 0:

                            # define if the factor takes a list as element
                            item['experimental_factors'][i]['is_list'] = \
                                fac_yaml[0]['list']

                            # set the ident_key from the general structure
                            # if special_case group was defined
                            if 'special_case' in fac_yaml[0] and 'group' in \
                                    fac_yaml[0]['special_case']:
                                item['experimental_factors'][i]['values'][
                                    'ident_key'] = fac_yaml[0]['special_case'][
                                    'group']

                    # copy the underlying structure (new_yaml) and remove the
                    # keys 'organism' and 'experimental_factors'
                    new_yaml = copy.deepcopy(key_yaml)
                    new_yaml['value'].pop('organism')
                    new_yaml['value'].pop('experimental_factors')

                    # redo the conditions
                    item = generate_metafile.get_redo_value(project_id,
                        new_yaml, 'experimental_setting',
                        not key_yaml['mandatory'], mandatory_mode,
                        copy.deepcopy(item), True, False, False, mode, key_yaml)[0]

                # no special case -> used for all the other keys
                else:

                    # key was already filled out
                    if key in item:

                        # call this function to edit the value of the key
                        item[key] = edit_item(project_id,
                            key, item[key], key_yaml['value'][key], full_yaml,
                            result_dict, mandatory_mode, mode)

                    # key was not filled out yet
                    else:

                        # call function to input information
                        item[key] = generate_metafile.get_redo_value(project_id,
                            key_yaml['value'][key], key, True, mandatory_mode,
                            result_dict, True, False, True, mode, key_yaml)

    # item is a single value
    else:

        # call function to input value
        item = generate_metafile.parse_input_value(
            key_name, key_yaml['desc'], key_yaml['whitelist'],
            key_yaml['input_type'], item)

    return item



import src.utils as utils
import src.web_interface.yaml_to_wi_object as yto
import src.web_interface.factors_and_conditions as fac_cond
import src.web_interface.wi_utils as wi_utils
import copy

disabled_fields = []

# TODO: code refactoring + documentation


def edit_wi_object(path, key_yaml, read_in_whitelists):
    """
    This function fills an empty wi object with the information of a metadata
    file
    :param key_yaml:
    :param path: the path containing the metadata file
    :return: wi_object: the filled wi object
    """
    # TODO: as Parameter at get_info
    meta_yaml = utils.read_in_yaml(path)
    whitelist_object = {}
    
    if meta_yaml is not None:
        if 'path' in meta_yaml:
            meta_yaml.pop('path')
        empty_object = yto.get_empty_wi_object(key_yaml, read_in_whitelists)
        wi_object = {}
        wi_object['all_factors'], real_val = get_all_factors(meta_yaml,
                                                             key_yaml)
        wi_object['publication_records'] = get_publication_records(meta_yaml)
        for part in empty_object:
            if part not in ['all_factors', 'publication_records']:
                wi_object[part], whitelist_object = new_fill(
                    meta_yaml[part], empty_object[part], key_yaml[part],
                    whitelist_object, real_val, read_in_whitelists)

    else:
        wi_object = yto.get_empty_wi_object(key_yaml, read_in_whitelists)

    wi_object['whitelists'] = whitelist_object
    wi_object = get_filenames(wi_object, meta_yaml)
    return wi_object


def get_publication_records(meta_yaml):
    publications = list(utils.find_keys(meta_yaml, 'publication'))
    publication_records = {}
    for elem in publications:
        for pub in elem:
            if pub['pubmed_id'] not in publication_records:
                publication_records[pub['pubmed_id']] = pub
    return publication_records


def get_filenames(wi_object, meta_yaml):
    old_filenames = {}
    old_sample_names = {}
    settings = list(utils.find_keys(meta_yaml, 'experimental_setting'))
    if len(settings) > 0:
        for setting in settings:
            for stg in setting:
                if 'setting_id' in stg:
                    setting_id = stg['setting_id']
                    bio_reps = list(utils.find_keys(stg, 'samples'))
                    if len(bio_reps) > 0:
                        for elem in bio_reps:
                            for sample in elem:
                                if 'sample_name' in sample:
                                    key = f'{setting_id}_{sample["sample_name"]}'
                                    tech_rep = list(utils.find_keys(sample, 'technical_replicates'))
                                    if len(tech_rep) > 0:
                                        if 'filenames' in tech_rep[0]:
                                            old_filenames[key] = tech_rep[0]['filenames']
                                        if 'sample_name' in tech_rep[0]:
                                            old_sample_names[key] = tech_rep[0]['sample_name']
    wi_object['old_sample_names'] = old_sample_names
    wi_object['old_filenames'] = old_filenames
    return wi_object


def new_fill(meta_yaml, wi_object, key_yaml, whitelist_object, real_val, read_in_whitelists):

    if isinstance(meta_yaml, dict):

        if 'headers' in wi_object:
            fill_key = 'value'
            filled_value = ''
            for header in wi_object['headers'].split(' '):
                filled_value = filled_value + ' ' + meta_yaml[header]
            filled_value = filled_value.lstrip(' ').rstrip(' ')
        else:
            if wi_object['position'].split(':')[-1] == 'experimental_setting':
                fill_key = 'input_fields'
                filled_value, whitelist_object = fill_experimental_setting(
                    wi_object, meta_yaml, key_yaml['value'], whitelist_object, real_val, read_in_whitelists)
            else:
                fill_key = 'input_fields'
                filled_value = copy.deepcopy(wi_object['input_fields'])
                for i in range(len(filled_value)):
                    if filled_value[i]['position'].split(':')[-1] in meta_yaml:
                        filled_value[i], whitelist_object = new_fill(
                            meta_yaml[filled_value[i]['position'].split(
                                ':')[-1]],
                            filled_value[i], key_yaml['value'][filled_value[i]['position'].split(
                                ':')[-1]], whitelist_object,
                            real_val, read_in_whitelists)

    elif isinstance(meta_yaml, list):
        fill_key = 'list_value'
        filled_value = []
        for i in range(len(meta_yaml)):
            f_val, whitelist_object = new_fill(
                meta_yaml[i], copy.deepcopy(wi_object), key_yaml,
                whitelist_object, real_val, read_in_whitelists)

            # TODO: WTF?
            if 'input_fields' in f_val:
                f_val = f_val['input_fields']
                if 'position' in f_val and f_val['position'].split(':')[-1] \
                        == 'experimental_setting':
                    f_val['input_disabled'] = True
            else:
                f_val = f_val['value']
            filled_value.append(f_val)

    else:
        fill_key = 'value'
        if wi_object['input_type'] == 'date':
            filled_value = wi_utils.str_to_date(meta_yaml)
        else:
            filled_value = meta_yaml

    if 'input_type' in wi_object and wi_object['input_type'] == \
            'single_autofill':
        fill_key = 'list_value'

    wi_object[fill_key] = filled_value

    if 'special_case' in key_yaml and 'edit' in key_yaml['special_case']:
        if key_yaml['special_case']['edit'] == 'not editable':
            wi_object['input_disabled'] = True
            wi_object['delete_disabled'] = True
        elif key_yaml['special_case']['edit'] == 'not removable':
            wi_object['delete_disabled'] = True
            if key_yaml['list']:
                wi_object['fixed_length'] = len(filled_value) if \
                    filled_value is not None else None
                if 'input_type' in wi_object and wi_object['input_type'] in ['select', 'group_select']:
                    wi_object['non_editable_val'] = filled_value
    return wi_object, whitelist_object


def fill_experimental_setting(wi_object, meta_yaml, key_yaml, whitelist_object,
                              real_val, read_in_whitelists):
    organism = ''
    filled_object = []
    for j in range(len(wi_object['input_fields'])):
        f = copy.deepcopy(wi_object['input_fields'][j])
        for key in meta_yaml:

            if wi_object['input_fields'][j]['position'].split(':')[-1] == key:
                if key == 'experimental_factors':
                    pass
                elif key == 'conditions':
                    sample_keys = list(utils.find_keys(key_yaml, 'samples'))
                    if len(sample_keys) > 0:

                        sample, whitelists = yto.parse_empty(
                            sample_keys[0],
                            'experimental_setting:conditions:biological_'
                            'replicates:samples', key_yaml,
                            {'organism': organism}, read_in_whitelists,
                            get_whitelist_object=True)
                        sample = sample['input_fields']

                        conditions = []

                        for cond in meta_yaml[key]:
                            samples = []
                            split_cond = utils.split_cond(
                                cond['condition_name'])
                            sample_name = utils.get_short_name(
                                cond['condition_name'], {}, key_yaml, read_in_whitelists=read_in_whitelists)
                            input_fields = fac_cond.get_samples(
                                split_cond, copy.deepcopy(sample), real_val,
                                key_yaml, sample_name, organism, read_in_whitelists)
                            if 'samples' in cond['biological_replicates']:
                                for s in cond['biological_replicates']['samples']:
                                    filled_keys = []
                                    for k in s:
                                        if s[k] is not None:
                                            # TODO: dict (disease usw.)
                                            # TODO: real_val
                                            if isinstance(s[k], list):
                                                for elem in s[k]:
                                                    if (k, elem) not in \
                                                            filled_keys and \
                                                            (k, elem) not in \
                                                            split_cond:
                                                        filled_keys.append(
                                                            (k, elem))
                                            elif (k, s[k]) not in split_cond:
                                                filled_keys.append((k, s[k]))
                                    sample_index = int(s["sample_name"].split(
                                        "_")[-1].replace("b", ""))
                                    cond_sample_name = f'{sample_name}_' \
                                                       f'{sample_index}'
                                    filled_sample = copy.deepcopy(input_fields)
                                    filled_sample = fac_cond.get_samples(
                                            filled_keys,
                                            filled_sample, real_val,
                                            key_yaml, cond_sample_name,
                                            organism, read_in_whitelists,
                                            is_factor=False)
                                    samples.append(copy.deepcopy(filled_sample))
                            title, readd = fac_cond.get_condition_title(
                                split_cond)
                            d = {'correct_value': cond['condition_name'],
                                 'title': title,
                                 'readd': readd,
                                 'input_disabled': False,
                                 'position': 'experimental_setting:condition',
                                 'list': True, 'mandatory': True,
                                 'list_value': samples,
                                 'desc': '',
                                 'input_fields': input_fields}
                            if 'special_case' in sample_keys[0] and 'edit' in \
                                    sample_keys[0]['special_case']:
                                if sample_keys[0]['special_case'][
                                    'edit'] == 'not editable':
                                    d['input_disabled'] = True
                                    d['delete_disabled'] = True
                                elif sample_keys[0]['special_case'][
                                    'edit'] == 'not removable':
                                    d['delete_disabled'] = True
                                    if sample_keys[0]['list']:
                                        d['fixed_length'] = len(
                                            samples) if \
                                            samples is not None else None
                            conditions.append(d)
                        # TODO: not editable and not removable for conditions?
                        f['list_value'] = conditions
                        whitelist_object[organism] = whitelists

                else:

                    if 'headers' in f and isinstance(meta_yaml[key], dict):
                        new_val = parse_headers(meta_yaml[key], f['headers'])
                    else:
                        new_val = meta_yaml[key]

                    if key == 'organism':
                        organism = new_val

                    if 'whitelist_keys' in f:
                        new_val = parse_whitelist_keys(
                            meta_yaml[key], f['whitelist_keys'],
                            utils.get_whitelist(key, {'organism': organism}, whitelist_object=read_in_whitelists))

                    if 'list' in f and f['list']:
                        f['list_value'] = new_val
                    else:
                        f['value'] = new_val
                filled_object.append(f)
    return filled_object, whitelist_object


def parse_headers(value, headers):

    if isinstance(headers, dict):
        header = None
        for k in headers:
            if sorted(headers[k].split(' ')) == sorted(list(value.keys())):
                header = headers[k].split(' ')
                break
    else:
        header = headers.split(' ')

    if header is not None:
        val = ''
        for h in header:
            val = f'{val}{" " if val != "" else ""}{value[h]}'
        value = val

    return value


def parse_whitelist_keys(value, whitelist_keys, whitelist):

    for key in whitelist_keys:
        if f'{value} ({key})' in whitelist:
            value = f'{value} ({key})'
            break

    return value


# TODO: value unit?
def get_all_factors(meta_yaml, real_val):
    """
    This function creates an object containing all experimental factors from a
    metadata yaml to be stored in a wi object
    :param real_val:
    :param meta_yaml: the read in metadata file
    :return: all_factors: an object containing all experimental factors
    """
    all_factors = []
    for setting in meta_yaml['experimental_setting']:
        setting_factors = []
        for factors in setting['experimental_factors']:
            setting_fac = {'factor': factors['factor']}
            w = utils.get_whitelist(factors['factor'], setting)
            setting_fac['values'] = []
            for elem in factors['values']:
                value = elem
                if w and 'headers' in w:
                    if 'headers' not in setting_fac:
                        setting_fac['headers'] = w['headers']
                    if isinstance(elem, dict):
                        value = parse_headers(value, w['headers'])

                if w and 'whitelist_keys' in w:
                    if 'whitelist_keys' not in setting_fac:
                        setting_fac['whitelist_keys'] = w['whitelist_keys']
                    if 'whitelist_type' in w and w['whitelist_type'] == \
                            'plain_group':
                        value = parse_whitelist_keys(
                            value, w['whitelist_keys'], w['whitelist'])

                if isinstance(elem, dict) and len(list(elem.keys())) == 2 and \
                        'unit' in elem and 'value' in elem:
                    value = f'{elem["value"]}{elem["unit"]}'

                if value != elem:
                    if isinstance(elem, dict):
                        # rewrite the value into a string
                        val = "|".join(
                            [f'{key}:"{elem[key]}"' for key in elem])
                        val = f'{factors["factor"]}:{"{"}{val}{"}"}'
                        real_val[val] = value
                    else:
                        real_val[elem] = value
                # TODO: nach oben rücken + rekursiv (eigene Funktion)
                elif isinstance(factors['values'], dict):
                    value = {value: factors['values'][value]}
                setting_fac['values'].append(value)
            setting_factors.append(setting_fac)
        all_factors.append(setting_factors)
    return all_factors, real_val



import plotly.graph_objects as go
import src.utils as utils
from src.heatmap.dash_utils import *
import plotly.express as px
import os
import base64
import io
from PIL import Image
import numpy


def get_heatmap(path, keys_yaml, show_setting_id=True, only_factors=False, mode='samples', labels='factors', background=False, sample_labels=False, condition_labels=False, transpose=False, drop_defaults=False):
    settings, experimental_factors, organisms, max_vals, options_pretty, annotated_dict, max_annotation, conditions = get_data(path, keys_yaml, mode=mode, drop_defaults=drop_defaults, transpose=transpose)

    techniques = list(utils.find_keys(path, 'techniques'))
    setting_techniques = {}
    if len(techniques) > 0:
        techniques = techniques[0]
        for elem in techniques:
            try:
                sett_id = elem['setting']
                try:
                    technique = elem['technique']
                    setting_techniques[sett_id] = technique
                except KeyError:
                    pass
            except KeyError:
                pass

    heatmaps=[]

    for value in settings:
        fig = None
        if len(settings[value]) > 0:

            options = [key.replace('_num', '') for key in settings[value] if key.endswith('_num')]
            
            if only_factors:
                sorter = experimental_factors[value]
            else:
                sorter = experimental_factors[value] + [o for o in options if o not in experimental_factors[value]]
            df_empty = settings[value].dropna(axis=1, how='all')
            sorter = [x for x in sorter if x in df_empty.columns]
            val_sorter = [0]
            for elem in sorter:
                if elem in max_annotation[value]:
                    val_sorter.append(max(val_sorter[-1]+(50 if not transpose else 150), val_sorter[-1]+(15*max_annotation[value][elem])+20))
                else:
                    val_sorter.append(val_sorter[-1]+(50 if not transpose else 150))

            df = [settings[value][f'{key}_num'] for key in sorter]   
            annotated = [annotated_dict[value][key] for key in sorter if key in annotated_dict[value]]
            if transpose:
                annotated = numpy.transpose(annotated)
            option_text = []
            label_text = []
            for key in sorter:
                if labels == 'all' or (labels == 'factors' and key in annotated_dict[value] and key in experimental_factors[value]):
                    label_vals = []
                    for elem in annotated_dict[value][key]:
                        splitted = elem.split('<br>')
                        new_val = []
                        for val in splitted:
                            if len(val) > 20:
                                new_val.append(f'{val[:18]}...')
                            else:
                                new_val.append(val)
                        label_vals.append('<br>'.join(new_val))
                    label_text.append(label_vals)
                else:
                    label_text.append(['']* len(annotated_dict[value][key]))

            for option in sorter:
                if option in experimental_factors[value]:
                    option_text.append(color("red", f'<b>{options_pretty[option]}</b>'))
                else:
                    option_text.append(color("black", options_pretty[option]))
                    
            plotly_colors = []

            colors_scales = [px.colors.qualitative.G10, px.colors.qualitative.Plotly, px.colors.qualitative.D3, px.colors.qualitative.T10]
            
            i = 0
            while len(plotly_colors) < max_vals[value]:
                plotly_colors += [x for x in colors_scales[i%len(colors_scales)] if x not in ["#DC3912", "#EF553B", "#D62728", "#F58518"]]
                i += 1

            colors = [[0, 'white']]
            for i in range(1, max_vals[value]+1):
                colors.append([i*1/(max_vals[value]), plotly_colors[i-1]])
                if i*1/(max_vals[value]) != 1:
                    colors.append([((i*1)/(max_vals[value]))+0.001, 'white'])

            if transpose:
                if mode == 'samples':
                    if condition_labels:
                        if sample_labels:
                            
                            y_value = [settings[value]['condition_label'], settings[value]['sample_label']]
                        else:
                            y_value = [settings[value]['condition_label'], settings[value]['sample_index']]
                    else:
                        if sample_labels:
                            y_value = [settings[value]['condition_index'], settings[value]['sample_label']]
                        else:
                            y_value = [settings[value]['condition_index'], settings[value]['sample_index']]
                else:
                    if condition_labels:
                        y_value = settings[value]['condition_label']
                    else:
                        y_value = settings[value]['condition_index']
                        
                x_value = val_sorter
                label_text = numpy.transpose(label_text)
                df = numpy.transpose(df)
            else:
                if mode == 'samples':
                    if condition_labels:
                        if sample_labels:
                            x_value = [settings[value]['condition_label'], settings[value]['sample_label']]
                        else:
                            x_value = [settings[value]['condition_label'], settings[value]['sample_index']]
                    else:
                        if sample_labels:
                            x_value = [settings[value]['condition_index'], settings[value]['sample_label']]
                        else:
                            x_value = [settings[value]['condition_index'], settings[value]['sample_index']]
                else:
                    if condition_labels:
                         x_value = settings[value]['condition_label']
                    else:
                        x_value = settings[value]['condition_index']
                y_value = val_sorter
                                
            heatmap = [go.Heatmap(
                        z=df,
                        zmin=0,
                        zmax=max_vals[value],
                        x=x_value,
                        y=y_value,
                        showscale=False,
                        customdata=annotated,
                        text=label_text,
                        texttemplate="%{text}",
                        hovertemplate = "%{customdata}",
                        hoverongaps = False,
                        colorscale = colors,
                        ),
                    ]
            
            data_input = heatmap
            
            if mode == 'samples':
                if sample_labels:
                    sample_label_height = max(settings[value]['sample_label_height'])
                else:
                    sample_label_height = 1
            else: sample_label_height = 0

            if condition_labels:
                condition_label_height = max(settings[value]['condition_label_height'])
            else:
                condition_label_height = 1

            my_cell_width = 150
            #my_cell_height = max(50, (15*max_annotation[value])+20)
            if transpose:
                top_margin = 100
                bottom_margin = 20
                left_margin = (150 if condition_labels else 100) + (150 if sample_labels else 100)
                right_margin = 20
                my_width = val_sorter[-1]
                my_height = 50 * (len(settings[value]["sample_index"]) if mode=='samples' else len(settings[value]["condition_index"]))
            else:
                top_margin = 15*sample_label_height + 15*condition_label_height + 70 # minimum 100
                bottom_margin = 20
                left_margin = 200
                right_margin = 20
                my_height = val_sorter[-1] #my_cell_height*len(sorter)

                if mode=='samples':
                    my_width = my_cell_width*len(settings[value]["sample_index"])
                else:
                    my_width = my_cell_width*len(settings[value]['condition_index'])
            organism_path = os.path.join(os.path.dirname(__file__), '..', '..', 'images', f'{organisms[value]}.png')
            images = None

            if os.path.isfile(organism_path):
                plotly_logo = base64.b64encode(open(organism_path, 'rb').read())
                imgdata = base64.b64decode(plotly_logo)
                im = Image.open(io.BytesIO(imgdata))
                im_width, im_height = im.size
                y_side = 100
                my_ysize=y_side/my_height
                x_side = y_side*im_width / im_height
                my_xsize = x_side/my_width

                left_margin = max(left_margin, x_side)

                images = [dict(
                    source='data:image/png;base64,{}'.format(plotly_logo.decode()),
                    xref="paper", yref="paper",
                    x=0, y=1,
                    sizey=my_ysize, sizex=my_xsize,
                    xanchor="right", yanchor="bottom"
            )]

            between = []
            technique_list = setting_techniques[value] if value in setting_techniques else None
            if technique_list is not None:
                technique_list = f' - {", ".join(technique_list)}'
            else:
                technique_list = ''
            for i in range(len(val_sorter)-1):
                between.append((val_sorter[i+1]+val_sorter[i])/2)
            layout = go.Layout(
                images = images,
                height=my_height + top_margin + bottom_margin,
                width=my_width + left_margin + right_margin,
                margin=dict(l=left_margin, r=right_margin, t=top_margin, b=bottom_margin),
                autosize=False,
                title=dict(
                    text=f'<b>Setting {value}{technique_list}</b>' if show_setting_id else '', 
                    font=dict(size=15), 
                    automargin=False,
                    yref='container',
                    x=(left_margin+(0.1*150))/(my_width+left_margin+right_margin),
                    y=1-(30/(my_height+top_margin+bottom_margin)*100)/(my_height+top_margin+bottom_margin),
                    xanchor="left", yanchor="top",
                    subtitle=dict(
                        text=f'Organism: {organisms[value]}',
                        font=dict(size=14, lineposition='under'),
                        ),
                    ),
                plot_bgcolor='rgb(255, 255, 255)' if background else 'rgba(0,0,0,0)',
                paper_bgcolor='rgb(255, 255, 255)' if background else 'rgba(0,0,0,0)',
                xaxis=dict(
                    side="top",
                    automargin=False,
                ) if not transpose else dict(
                    tickmode="array",
                    ticktext=option_text, 
                    tickvals=between,
                    side="top",
                    automargin=False
                ),
                yaxis=dict(
                    tickmode="array",
                    ticktext=option_text, 
                    tickvals=between,
                    autorange="reversed",
                    automargin=False
                ) if not transpose else dict(
                    automargin=False,
                    autorange="reversed",
                )
            )    

            fig = go.Figure(data=data_input, layout=layout)     
            if transpose:
                for i in val_sorter:
                    fig.add_vline(i, line_width=0.5)
                
                for i in between:
                    fig.add_vline(i, line_dash="dash", line_color='lightgrey',layer='below') 

                if mode == 'samples':
                    for i in range(len(settings[value]['sample_index'])+1):
                        fig.add_hline(i-0.5, line_width=0.5)
                else:
                    for i in range(len(settings[value]['condition_index'])+1):
                        fig.add_hline(i-0.5, line_width=0.5)
                
                fig.add_vrect(x0=0, x1=val_sorter[len(experimental_factors[value])], line=dict(color="red", width=5), layer='above')

            else:
                for i in val_sorter:
                    fig.add_hline(i, line_width=0.5)
                
                for i in between:
                    fig.add_hline(i, line_dash="dash", line_color='lightgrey',layer='below')      

                if mode == 'samples':
                    for i in range(len(settings[value]['sample_index'])+1):
                        fig.add_vline(i-0.5, line_width=0.5)
                else:
                    for i in range(len(settings[value]['condition_index'])+1):
                        fig.add_vline(i-0.5, line_width=0.5)
                
                fig.add_hrect(y0=0, y1=val_sorter[len(experimental_factors[value])], line=dict(color="red", width=5), layer='above')
            
            fig.update_layout(modebar={'remove':['zoom', 'pan', 'zoomIn', 'zoomOut', 'autoScale']})
            fig.layout.yaxis.fixedrange = True

        heatmaps.append((value, fig, conditions[value] if len(list(conditions[value].keys())) > 0 else None))
    return heatmaps


def color(color, text):
    return f"<span style='color:{str(color)}'> {str(text)} </span>"




import os
from src import utils
import datetime
from tabulate import tabulate
import readline
import copy
import re


class WhitelistCompleter:
    def __init__(self, whitelist):
        self.whitelist = whitelist

    def complete(self, text, state):
        results = [x for x in self.whitelist if
                   text.lower() in x.lower()] + [None]
        if len(results) > 30:
            results = results[:30]
        return results[state]


class Input:

    def __init__(self, path, project_id, mandatory_only, filename, key_yaml, email):
        self.path = path
        self.project_id = project_id
        self.mandatory_only = mandatory_only
        self.filename = filename
        self.key_yaml = key_yaml
        self.result_dict = {}
        self.size = os.get_terminal_size().columns
        self.conditions = {}
        self.publications = {}
        self.generate_end = []
        self.email = email
        self.setting_ids = []

    def parse_input_value(self, key, structure, allow_float=False):
        """
        This function lets the user enter an input and tests if the input is
        valid.
        :param key: the key that should be filled
        :param structure: dictionary containing information about the key
        """

        # read in whitelist if the key has one or set the whitelist to None
        if structure['whitelist']:
            whitelist = utils.get_whitelist(key, self.result_dict)
        else:
            whitelist = None

        # test if a whitelist is defined
        if whitelist is not None:

            # test if the whitelist is a dictionary
            if isinstance(whitelist['whitelist'], dict):

                # rewrite the whitelist to a list
                w = [x for xs in list(whitelist['whitelist'].values()) if xs is
                     not None for x in xs]

                # use autocomplete if the whitelist > 30 values
                if len(w) > 30:
                    # TODO: print_desc Funktion
                    print(f'\n---\n{structure["desc"]}')
                    input_value = self.complete_input(w, key)

                    # repeat if the input value does not match the whitelist
                    while input_value not in w:
                        print(f'The value you entered does not match the '
                              f'whitelist. Try tab for autocomplete.')
                        input_value = self.complete_input(w, key)

                else:

                    # parse grouped whitelist if the whitelist < 30 values
                    # TODO: description for experimental factors + factor desc
                    input_value = self.parse_group_choose_one(
                        whitelist['whitelist'], w, f'{key}:')

            # use autocomplete if the whitelist > 30 values
            elif len(whitelist['whitelist']) > 30:
                # TODO: print_desc + remove duplicate -> own function
                print(f'\n---\n{structure["desc"]}')
                input_value = self.complete_input(whitelist['whitelist'], key)

                # repeat if the input value does not match the whitelist
                while input_value not in whitelist['whitelist']:
                    print(f'The value you entered does not match the '
                          f'whitelist. Try tab for autocomplete.')
                    input_value = self.complete_input(whitelist['whitelist'], key)

            # print indexed whitelist and parse user input
            else:
                # TODO: print_desc
                print(f'\n---\n{structure["desc"]}')
                input_value = self.parse_list_choose_one(
                    whitelist['whitelist'], f'{key}:')

            # TODO: remove w_key?
            w_key = None
            if whitelist['whitelist_type'] == 'plain_group':
                for k in whitelist['whitelist_keys']:
                    if input_value.endswith(f' ({k})'):
                        input_value = input_value.replace(f' ({k})', '')
                        w_key = k

            # test if the whitelist contains the key 'headers'
            if 'headers' in whitelist:
                if whitelist['whitelist_type'] == 'group' or whitelist[
                        'whitelist_type'] == 'plain_group' and w_key is not None:
                    if w_key in whitelist['headers']:
                        headers = whitelist['headers'][w_key].split(' ')
                        vals = input_value.split(' ')

                        # create a dictionary to store the new value
                        value = {}

                        # iterate through the headers and save the header and value of the
                        # same index into a dictionary with header as key
                        for i in range(len(headers)):
                            value[headers[i]] = vals[i]

                        # overwrite the input value with the dictionary
                        input_value = value

                else:

                    # split the headers and the input value at ' ' and save each to
                    # a list
                    headers = whitelist['headers'].split(' ')
                    vals = input_value.split(' ')

                    # create a dictionary to store the new value
                    value = {}

                    # iterate through the headers and save the header and value of the
                    # same index into a dictionary with header as key
                    for i in range(len(headers)):
                        value[headers[i]] = vals[i]

                    # overwrite the input value with the dictionary
                    input_value = value

        # no whitelist
        else:

            # print a description if one is stated in the metadata structure
            # TODO: print_desc
            if structure["desc"] != '':
                print(f'\n---\n{structure["desc"]}\n')

            # input type boolean
            if structure['input_type'] == 'bool':
                input_value = self.parse_list_choose_one(
                    ['True ', 'False '], f'Is the sample {key}')

            else:
                
                # print the key, add a newline if a description was printed
                if structure["desc"] == '':
                    input_value = input(f'\n{key}: ')
                else:
                    input_value = input(f'{key}: ')
                # no user input -> repeat
                
                if input_value == '':
                    print(f'Please enter something.')
                    input_value = self.parse_input_value(key, structure, allow_float=allow_float)

                # input type number
                # TODO: allow float
                if structure['input_type'] == 'number':
                    try:
                        input_value = int(input_value)
                    except ValueError:
                        if allow_float:
                            try:
                                input_value = float(input_value)
                            except ValueError:
                                print(f'Input must be of type int or float. Try again.')
                                input_value = self.parse_input_value(key, structure, allow_float=allow_float)
                        else:
                            print(f'Input must be of type int. Try again.')
                            input_value = self.parse_input_value(key, structure, allow_float=allow_float)

                # input type date (format dd.mm.yyyy)
                elif structure['input_type'] == 'date':
                    try:
                        input_date = input_value.split('.')
                        if len(input_date) != 3 or len(
                                input_date[0]) != 2 or len(
                            input_date[1]) != 2 or len(input_date[2]) != 4:
                            raise SyntaxError
                        input_value = datetime.date(int(input_date[2]),
                                                    int(input_date[1]),
                                                    int(input_date[0]))
                        input_value = input_value.strftime("%d.%m.%Y")
                    except (IndexError, ValueError, SyntaxError) as e:
                        print(f'Input must be of type \'DD.MM.YYYY\'.')
                        input_value = self.parse_input_value(key, structure, allow_float=allow_float)

                else:

                    # input type restricted_short_text
                    if structure['input_type'] == 'restricted_short_text':
                        if 'special_case' in structure and 'restriction' in structure['special_case']:
                            if 'max_length' in structure['special_case']['restriction'] and len(input_value) > structure['special_case']['restriction']['max_length']:
                                print(f'Input must not exceed {structure["special_case"]["restriction"]["max_length"]} characters. Please try again.')
                                input_value = self.parse_input_value(key, structure, allow_float=allow_float)
                            elif 'regex' in structure['special_case']['restriction']:
                                pattern = re.compile(structure['special_case']['restriction']['regex'])
                                if pattern.match(input_value):
                                    print(f'Input does not conform to defined pattern. Please try again.')
                                    input_value = self.parse_input_value(key,
                                                                         structure, allow_float=allow_float)

                    # invalid character
                    if '\"' in input_value:
                        print(f'Invalid symbol \'\"\'. Please try again.')
                        input_value = self.parse_input_value(key, structure, allow_float=allow_float)
                    elif '{' in input_value:
                        print(f'Invalid symbol \'{"{"}\'. Please try again.')
                        input_value = self.parse_input_value(key, structure, allow_float=allow_float)
                    elif '}' in input_value:
                        print(f'Invalid symbol \'{"}"}\'. Please try again.')
                        input_value = self.parse_input_value(key, structure, allow_float=allow_float)
                    elif '|' in input_value:
                        print(f'Invalid symbol \'|\'. Please try again.')
                        input_value = self.parse_input_value(key, structure, allow_float=allow_float)

        # return the user input
        return input_value

    def parse_list_choose_one(self, whitelist, header):
        """
        This function prints an indexed whitelist and prompts the user to
        choose a value by specifying the index.
        :param whitelist: a list of values for the user to choose from
        :param header: a headline or description that should be printed
        :return: value: the whitelist value chosen by the user
        """

        # print the given header and indexed whitelist and parse the user input
        try:
            print(f'{header}\n')
            self.print_option_list(whitelist, False)
            value = whitelist[int(input()) - 1]

        # redo the input prompt if the user input is not an integer
        except (IndexError, ValueError):
            print(f'Invalid entry. Please enter a number between 1 and '
                  f'{len(whitelist)}')
            value = self.parse_list_choose_one(whitelist, header)

        # TODO: direkt bool benutzen?
        if value == 'True ':
            value = True
        elif value == 'False ':
            value = False

        # return the user input
        return value

    def parse_group_choose_one(self, whitelist, w, header):
        """
        This function prints a grouped whitelist with indexes and lets the user
        choose one value.
        :param whitelist: the grouped whitelist as a dictionary
        :param w: a list containing all whitelist values
        :param header: a header that should be printed
        :return:
        """

        try:

            # print a header or description
            print(f'{header}\n')

            # set index to 1
            i = 1

            # iterate through keys in whitelist
            for key in whitelist:

                # print the indexed whitelist values with the keys as captions
                if whitelist[key] is not None:
                    print(f'\033[1m{key}\033[0m')
                    for value in whitelist[key]:
                        print(f'{i}: {value}')
                        i += 1

            # select the whitelist value that fits the input index
            value = w[int(input()) - 1]

        # redo the input if the passed value is not an integer >1
        except (IndexError, ValueError):
            print(f'Invalid entry. Please enter a number between 1 and '
                  f'{len(whitelist)}')
            value = self.parse_list_choose_one(whitelist, header)

        # return the user input
        return value

    def get_input_list(self, node, item):
        """
        This function asks the user to enter a list of values divided by comma and
        parses the input.
        :param node: a part of the keys.yaml
        :param item: the key that should be filled
        :return: used_values: the filled in list of values
        """
        # test if a whitelist exists for the item or if the special case 'merge'
        # was defined (means that the input is treated like a single value and then
        # split into a dictionary, e.g. gene -> gene_name, ensembl_id)
        if 'whitelist' in node and node['whitelist'] or 'special_case' in node \
                and 'merge' in node['special_case']:

            # read in whitelist
            whitelist = utils.get_whitelist(item, self.result_dict)

            # test if the whitelist is not None
            if whitelist is not None:

                # test if autocompletion is needed (for whitelists longer than 30)
                if len(whitelist['whitelist']) > 30:

                    # define an empty list to store the input values
                    used_values = []

                    # set parameter redo to True to initiate an input loop
                    redo = True

                    # request user input
                    # TODO: explain autocomplete
                    print(f'\nPlease enter the values for experimental factor '
                          f'{item}.')

                    while redo:

                        # prompt user input via autocompletion
                        input_value = self.complete_input(whitelist['whitelist'],
                                                     item)

                        # test if the input matches the whitelist
                        if input_value in whitelist['whitelist']:

                            # add input to list
                            used_values.append(input_value)

                            # ask the user if he wants to input another item
                            redo = self.parse_list_choose_one(
                                ['True ', 'False '],
                                f'\nDo you want to add another {item}?')

                        else:

                            # print message for invalid value, loop is repeated
                            print(f'The value you entered does not match the '
                                  f'whitelist. Try tab for autocomplete.')

                else:

                    # test for different whitelist types
                    if whitelist['whitelist_type'] == 'plain' or \
                            whitelist['whitelist_type'] == 'plain_group':

                        # for plain whitelists print the whitelist with indices and
                        # parse the user input
                        self.print_option_list(whitelist['whitelist'], '')
                        used_values = self.parse_input_list(whitelist['whitelist'],
                                                       False)

                    elif whitelist['whitelist_type'] == 'group':

                        # rewrite the grouped whitelist of type dictionary into a
                        # plain list
                        w = [x for xs in list(whitelist['whitelist'].values())
                             for
                             x in xs]

                        # set an index i and start with 1
                        i = 1

                        # iterate over the grouped whitelist (dictionary)
                        for w_key in whitelist['whitelist']:

                            # print the key/group of the whitelist
                            print(f'\033[1m{w_key}\033[0m')

                            # iterate over the values within a group
                            for value in whitelist['whitelist'][w_key]:
                                # print the value with the current index
                                print(f'{i}: {value}')

                                # increase the index by 1
                                i += 1

                        # parse the user input (indices) and match them to the
                        # whitelist values using the plain list
                        used_values = self.parse_input_list(w, False)

                # HANDLE PLAIN GROUPED WHITELISTS:
                # a plain grouped whitelist is a grouped whitelist that contains
                # more than 30 values and is therefor to long to be displayed.
                # In order to use autocompletion on such a whitelist, it is
                # rewritten into a plain whitelist. In order to not loose the
                # groups, they are added in round braces to the end of the
                # whitelist values (e.g. value 'GFP' in group 'other' in the
                # enrichment whitelist turns into 'GFP (other)'). The following
                # code removes those groups from the values again and saves them
                # into a list so that they can still be accessed to handle headers
                # in the part below.

                # set an empty list to save keys of plain grouped whitelists
                w_keys = []

                # remove group keys from values if the whitelist is of type
                # 'plain_group'
                if whitelist['whitelist_type'] == 'plain_group':

                    # iterate over the input values
                    for i in range(len(used_values)):

                        # iterate over the keys of the plain group whitelist
                        for k in whitelist['whitelist_keys']:

                            # test if the input value contains the key in braces
                            # at the end
                            if used_values[i].endswith(f' ({k})'):
                                # remove the key from the end of the value
                                used_values[i] = used_values[i].replace(
                                    f' ({k})', '')

                                # add the key to the list of whitelist keys
                                w_keys.append(k)

                # HANDLE HEADERS:
                # Some whitelists contain headers. Those headers represent keys,
                # the input values should be split into. An example can be found in
                # the gene whitelist where the header is set to
                # 'gene_name ensemble_ID' leading to the value
                # 'TSPAN6 ENSG00000000003' being saved as {'gene_name': 'TSPAN6',
                # 'ensemble_ID': 'ENSG00000000003'}. Headers can also occur in
                # grouped whitelists where they are defined for every group
                # separately. In order to know how to split a value in a grouped
                # whitelists one needs to know which group the value belongs to.
                # For grouped whitelists that were rewritten to plain whitelists
                # (see part above) this information is saved in the list w_keys
                # that was filled above. The following splits values into a
                # dictionary according to their headers.

                # test if headers were defined in the whitelist
                if 'headers' in whitelist:

                    # test if whitelist is of type group or plain group and if
                    # w_keys were defined
                    # TODO: works for group?
                    if whitelist['whitelist_type'] == 'group' or \
                            whitelist['whitelist_type'] == 'plain_group' \
                            and len(w_keys) > 0:

                        # iterate over input values
                        for i in range(len(used_values)):

                            # look at index of value in w_keys and see if a header
                            # was defined for that w_key
                            if w_keys[i] in whitelist['headers']:

                                # TODO: own function for header?

                                # split the header at the whitespace to get a list
                                # of keys
                                headers = whitelist['headers'][
                                    w_keys[i]].split(
                                    ' ')

                                # split the value at whitespace to get the
                                # according values
                                vals = used_values[i].split(' ')

                                # initialize an empty dictionary to save the
                                # key-value-pairs to
                                used_values[i] = {}

                                # iterate over the keys in the header
                                for j in range(len(headers)):
                                    # save the key and value at the same index into
                                    # the dictionary
                                    used_values[i][headers[j]] = vals[j]

                    else:

                        # split the header of a non-group whitelist at whitespace
                        # to get a list of keys
                        headers = whitelist['headers'].split(' ')

                        # iterate over the input values
                        for i in range(len(used_values)):

                            # split the value at whitespace to get the
                            # according values
                            vals = used_values[i].split(' ')

                            # initialize an empty dictionary to save the
                            # key-value-pairs to
                            used_values[i] = {}

                            # iterate over the keys in the header
                            for j in range(len(headers)):
                                # save the key and value at the same index into
                                # the dictionary
                                used_values[i][headers[j]] = vals[j]

            else:
                # TODO: kann weg?
                print('No whitelist')
                used_values = [0]

        # no whitelist
        else:

            # get the value_type for the input from the underlying structure
            value_type = node['input_type']

            if 'text' in value_type:
                used_values = []
                redo = True
                while redo:
                    used_values.append(self.parse_input_value(item, node))
                    redo = self.parse_list_choose_one(
                        ['True ', 'False '],
                        f'\nDo you want to add another '
                        f'{item}?')
            else:
                # request user input
                print(
                    f'\nPlease enter a list of {value_type} values for key {item} '
                    f'divided by comma:\n')

                # parse user input
                used_values = self.parse_input_list(value_type, False)

        return used_values

    def print_option_list(self, options, desc):
        """
        This function prints an indexed whitelist.
        :param options: the whitelist values
        :param desc: a description to be printed
        """

        # test if a description was given
        if desc:

            # create a nested list with every sublist containing an index, option
            # and description
            data = [[f'{i + 1}:', f'{options[i]}', desc[i]] for i in
                    range(len(options))]
            # print the nested list as a table
            print(tabulate(data, tablefmt='plain',
                           maxcolwidths=[self.size * 1 / 8,
                                         self.size * 3 / 8,
                                         self.size * 4 / 8]))
        else:

            # create a nested list with every sublist containing only index and
            # option
            data = [[f'{i + 1}:', f'{options[i]}'] for i in
                    range(len(options))]

            # print the nested list as a table
            print(tabulate(data, tablefmt='plain',
                           maxcolwidths=[self.size * 1 / 8,
                                         self.size * 7 / 8]))

    def parse_input_list(self, options, terminable, allow_float=False):
        """
        This function parses the user input for a list.
        :param options: possible input options as a list or 'number' for
                        number input
        :param terminable: a bool to state if nothing can be input
        :return: input_list: a list containing the input values
        """

        # promt user input for a comma divided list
        input_list = input()

        # if terminable is set to True and the user inputs 'n', nothing is returned
        if terminable and input_list.lower() == 'n':
            return None

        else:

            # test if a list was submitted as options
            if isinstance(options, list):

                # split the input list of indices and try to match it with the
                # option list
                try:
                    input_list = [options[int(i.strip()) - 1] for i in
                                  input_list.split(',')]

                except (IndexError, ValueError):

                    # Call this function to redo the input if it does not match the
                    # option list
                    print(f'Invalid entry, try again:')
                    input_list = self.parse_input_list(options, terminable, allow_float=allow_float)

            else:

                try:
                    # split the input list at ','
                    input_list = [x.strip() for x in input_list.split(',')]

                    # convert the input list to integers if they are of type
                    # 'number'
                    if options == 'number':
                        for i in range(len(input_list)):
                            try:
                                input_list[i] = int(input_list[i])
                            except ValueError:
                                if allow_float:
                                    try:
                                        input_list[i] = float(input_list[i])
                                    except ValueError:
                                        print(
                                            f'Invalid entry. Please enter numbers divided '
                                            f'by comma.')
                                        input_list = self.parse_input_list(
                                            options, terminable, allow_float=allow_float)
                                else:
                                    print(
                                        f'Invalid entry. Please enter integers divided '
                                        f'by comma.')
                                    input_list = self.parse_input_list(options,
                                                                       terminable, allow_float=allow_float)

                except IndexError:

                    # Call this function to redo the input if the split or
                    # conversion does not work
                    print(f'Invalid entry. Please enter numbers divided '
                          f'by comma.')
                    input_list = self.parse_input_list(options, terminable, allow_float=allow_float)

        return input_list

    def complete_input(self, whitelist, key):
        """
        This function uses a completer to autofill user input and print matching
        values.
        :param whitelist: a whitelist with possible values
        :param key: the key that should be filled
        :return: input_value: the value that was input by the user
        """

        # TODO: Doku
        print(f'Press tab once for autofill if'
              f' possible or to get a list of up to'
              f' 30 possible input values.\n')
        readline.parse_and_bind("tab: complete")
        readline.parse_and_bind("set show-all-if-ambiguous On")
        readline.parse_and_bind("set show-all-if-unmodified On")
        completer = WhitelistCompleter(whitelist)
        readline.set_completer(completer.complete)
        readline.set_completer_delims('')
        input_value = input(f'{key}: ')
        readline.parse_and_bind('tab: self-insert')
        return input_value

    def get_value_unit(self, structure):
        """
        This function prompts the user to input a value_unit.
        :param result_dict: a dictionary containing filled information
        :return: val_un: a dictionary containing the unit and value
        """

        # create a dictionary containing 'unit' and 'value' and call the
        # 'parse_input_value' function to request those information from the user
        val_un = {'unit': self.parse_input_value('unit', structure['value']['unit']),
                  'value': self.parse_input_value('value', structure['value']['value'], allow_float=True)}

        return val_un

    def get_list_value_unit(self, ex_factor, structure):
        """
        This function prompts the user to enter a list of value_units for
        experimental factors.
        :param result_dict: a dictionary containing filled information
        :param ex_factor: the experimental factor that contains the value_unit
        :return: a dictionary containing a unit and a list of values
        """
        # print request for unit input
        print(f'\nPlease enter the unit for factor {ex_factor}:')
        # call function to input unit
        unit = self.parse_input_value('unit', structure['value']['unit'])

        # initialize empty list to store the values with units
        val_un = []

        # print request for value input
        print(f'\nPlease enter int values for factor {ex_factor} (in {unit}) '
              f'divided by comma:')

        # call function to input values
        value = self.parse_input_list('number', False, allow_float=True)

        # iterate through the input values and add a dictionary containing a value
        # and its unit to the 'val_un' list
        for val in value:
            val_un.append({'unit': unit, 'value': val})

        return val_un

    def merge_dicts(self, a, b):
        """
        This function merges two dictionaries with the same structure to create
        one.
        :param a: the first dictionary
        :param b: the second dictionary
        :return: res: the merged dictionary
        """

        # test if dictionary 'a' is a list
        if isinstance(a, list):

            # initialize a list to save combined content of dict 'a' and 'b'
            res = []

            # iterate over dict 'a'
            for i in range(len(a)):
                # call this function for every index of the list and add it to the
                # list
                res.append(self.merge_dicts(a[i], b[i]))

        # if dict 'a' is a dictionary
        elif isinstance(a, dict):

            # get a list of all keys of dict 'b'
            b_keys = list(b.keys())

            # initialize a dictionary to save the combined information to
            res = {}

            # iterate over the keys of dict 'a'
            for key in a.keys():

                # test if the key is in dict 'b'
                if key in b_keys:

                    # call this function to merge the values of the key of dict 'a'
                    # and 'b'
                    res[key] = self.merge_dicts(a[key], b[key])

                    # remove the key from the list of keys for dict 'b'
                    b_keys.remove(key)

                else:

                    # add the key and value from dict 'a' to the result
                    res[key] = a[key]

            # iterate over all keys left in the list of keys for dict 'b'
            for key in b_keys:
                # add the key and value from dict 'b' to the result
                res[key] = b[key]

        # single value
        else:

            # TODO: different lists?

            # set the value of a as the result
            res = a

        return res

    def get_combinations(self, values, key):
        """
        This function creates combinations for experimental factors that can occur
        multiple times in one condition and lets the user choose those that were
        analyzed.
        :param values: the possible values of the factor
        :param key: the name of the experimental factor
        :return: used_values: the combinations of the experimental factor that were
                              used in the conditions
        """
        if isinstance(values, dict):
            is_dict = True
        else:
            is_dict = False
        if 'ident_key' in values:
            if values['ident_key'] in values and len(
                    values[values['ident_key']]) > 1:
                multi = self.parse_list_choose_one(
                    ['True ', 'False '],
                    f'\nCan one sample contain multiple {key}s?')
            else:
                multi = False
                values.pop('ident_key')
        else:
            multi = self.parse_list_choose_one(
                ['True ', 'False '],
                f'\nCan one sample contain multiple {key}s?')

        merge_values = utils.get_combis(copy.deepcopy(values), key,
                                        self.result_dict, self.key_yaml)
        print(
                f'\nPlease select the analyzed combinations for {key} '
                f'(1-{len(merge_values)}) divided by comma:\n')
        self.print_option_list(merge_values, False)
        used_values = self.parse_input_list(merge_values, False)

        return used_values



import src.utils as utils
import os

# TODO: comment


def save_object(dictionary, path, filename, edit_state):
    """
    This function saves the yaml structure into a file
    :param dictionary: the parsed wi object in yaml format
    :param path: the path to save the file to
    :param filename: the name of the file
    :return: new_filename: the name under which the file was saved
    """
    project_id = list(utils.find_keys(dictionary, 'id'))
    if len(project_id) > 0:
        project_id = project_id[0]
    else:
        project_id = ''

    if not edit_state:
        filename = f'{project_id}_{filename}_metadata.yaml'

    utils.save_as_yaml(dictionary, os.path.join(path, filename))

    return filename, project_id


def save_filenames(file_str, path):
    """
    This function saves the generated filenames into a file
    :param file_str: the filenames to be saved
    :param path: the path to save the file to
    :return: filename: the name under which the generated filenames are saved
    """
    if file_str is not None:
        filename = f'{file_str[0]}_samples.txt'
        text_file = open(os.path.join(path, filename), "w")
        text_file.write(file_str[1])
        text_file.close()
    else:
        filename = None
    return filename



import copy
import sys

sys.path.append("metadata-organizer")
import os
import random
import string
import subprocess
import time

import src.utils as utils
import src.web_interface.editing as editing
import src.web_interface.factors_and_conditions as fac_cond
import src.web_interface.file_io as file_io
import src.git_whitelists as gwi
import src.web_interface.html_output as html_output
import src.web_interface.searching as searching
import src.web_interface.validation as validation
import src.web_interface.whitelist_parsing as whitelist_parsing
import src.web_interface.wi_object_to_yaml as oty
import src.web_interface.yaml_to_wi_object as yto
import src.heatmap.create_heatmap as create_heatmap
from jinja2 import Template

# This script contains all functions for generation of objects for the web
# interface


class Webinterface:

    def __init__(self, config):
        (
            self.whitelist_repo,
            self.whitelist_branch,
            self.whitelist_path,
            self.username,
            self.password,
            structure,
            self.update_whitelists,
            self.output_path,
            self.filename,
            self.email,
        ) = utils.parse_config(config)
        self.structure = utils.read_in_yaml(structure)
        self.whitelist_version = fetch_whitelists(self.__dict__)

    def to_dict(self):
        return self.__dict__


def fetch_whitelists(pgm_object):
    whitelist_version = gwi.get_whitelists(
        pgm_object["whitelist_path"],
        pgm_object["whitelist_repo"],
        pgm_object["whitelist_branch"],
        pgm_object["update_whitelists"],
    )
    return whitelist_version


def get_whitelist_object(pgm_object):
    whitelist_object = {
        "whitelists": whitelist_parsing.get_whitelist_object(pgm_object),
        "version": pgm_object["whitelist_version"],
    }
    return whitelist_object


def get_empty_wi_object(pgm_object, read_in_whitelists):
    return yto.get_empty_wi_object(pgm_object["structure"], read_in_whitelists)


def is_empty(pgm_object, wi_object, read_in_whitelsits):
    emtpy_object = yto.get_empty_wi_object(pgm_object["structure"], read_in_whitelsits)
    if wi_object == emtpy_object:
        empty = True
    else:
        empty = False
    return {"empty": empty, "object": emtpy_object}


def get_single_whitelist(ob, read_in_whitelists):
    return whitelist_parsing.get_single_whitelist(ob, read_in_whitelists)


def get_factors(pgm_object, organism, read_in_whitelists):
    return fac_cond.get_factors(organism, pgm_object["structure"], read_in_whitelists)


def get_conditions(pgm_object, factors, organism_name, read_in_whitelists):
    return fac_cond.get_conditions(
        factors, organism_name, pgm_object["structure"], read_in_whitelists
    )


def validate_object(pgm_object, wi_object, read_in_whitelists, finish=False):
    new_object = copy.deepcopy(wi_object)
    return validation.validate_object(
        new_object, pgm_object["structure"], read_in_whitelists, finish,
        pgm_object['email']
    )


def get_summary(pgm_object, wi_object, read_in_whitelists):
    return html_output.get_summary(
        wi_object, pgm_object["structure"], read_in_whitelists
    )


def save_object(dictionary, path, filename, edit_state):
    object, id = file_io.save_object(dictionary, path, filename, edit_state)
    return object, id


def save_filenames(file_str, path):
    return file_io.save_filenames(file_str, path)

def get_plot(pgm_object, config, path, project_id):
    uuid = "".join(
            random.choice(string.ascii_uppercase + string.digits) for _ in range(5)
        )
    filename = f"{uuid}_{time.time()}"
    working_path = os.path.join(os.path.dirname(__file__), "..", "..")
    proc = subprocess.Popen(
            [
                "python3",
                "metadata-organizer/metaTools.py",
                "find",
                "-p",
                path,
                "-s",
                f'project:id:"{project_id}',
                "-c",
                config,
                "-o",
                "json",
                "-f",
                filename,
                "-sv",
            ],
            cwd=working_path,
        )
    proc.wait()
    res = utils.read_in_json(os.path.join(working_path, f"{filename}.json"))
    os.remove(os.path.join(working_path, f"{filename}.json"))

    try:
        yaml_file = utils.read_in_yaml(res['data'][0]['path'])
        template = Template(
        '''              
        {% if input.html %}
            {{ input.html }}
        {% else %}            
            <div style="overflow:auto; overflow-y:hidden; margin:0 auto; white-space:nowrap; padding-top:20">
                    {% if input.plot %}
                        {{ input.plot }}
                    {% endif %}
                    
                    {% if input.missing_samples %}
                        <i>Conditions without samples:</i>
                        {{ input.missing_samples }}
                    {% endif %}
            </div>
        {% endif %}
        '''
        )
        plots = create_heatmap.get_heatmap(yaml_file, pgm_object['structure'], show_setting_id=False)
        plot_list = []
        for elem in plots:
            add_plot = {}
            if elem[1] is not None:
                add_plot['plot'] = elem[1]
            if elem[2] is not None:
                add_plot['missing_samples'] = html_output.object_to_html(elem[2], 0, False)
            plot_list.append({'title': elem[0], 'plot': template.render(input=add_plot)})
    except:
        plot_list = []
    return plot_list


# TODO: fix path
def get_meta_info(config, path, project_ids):

    if not isinstance(project_ids, list):
        project_ids = [project_ids]

    metafile = {}
    html_str = ""
    for project_id in project_ids:
        uuid = "".join(
            random.choice(string.ascii_uppercase + string.digits) for _ in range(5)
        )
        filename = f"{uuid}_{time.time()}"
        working_path = os.path.join(os.path.dirname(__file__), "..", "..")
        proc = subprocess.Popen(
            [
                "python3",
                "metadata-organizer/metaTools.py",
                "find",
                "-p",
                path,
                "-s",
                f'project:id:"{project_id}',
                "-c",
                config,
                "-o",
                "json",
                "-f",
                filename,
                "-sv",
            ],
            cwd=working_path,
        )
        proc.wait()
        res = utils.read_in_json(os.path.join(working_path, f"{filename}.json"))
        os.remove(os.path.join(working_path, f"{filename}.json"))

        html_str, metafile = searching.get_meta_info(
            html_str,
            res["data"],
            project_id,
            res["validation_reports"] if "validation_reports" in res else None,
        )

    if html_str == "":
        html_str = "No metadata found.<br>"
    return html_str, metafile


def get_search_mask(pgm_object):
    return searching.get_search_mask(pgm_object["structure"])


# TODO: fix path
def find_metadata(config, path, search_string):
    start = time.time()
    uuid = "".join(
        random.choice(string.ascii_uppercase + string.digits) for _ in range(5)
    )
    filename = f"{uuid}_{time.time()}"
    working_path = os.path.join(os.path.dirname(__file__), "..", "..")
    proc = subprocess.Popen(
        [
            "python3",
            "metadata-organizer/metaTools.py",
            "find",
            "-p",
            path,
            "-s",
            search_string,
            "-c",
            config,
            "-o",
            "json",
            "-f",
            filename,
            "-sv",
        ],
        cwd=working_path,
    )
    proc.wait()
    subprocess_end = time.time()
    print(
        f'Subprocess "FIND" took {"%.2f" % (subprocess_end - start)} seconds.')
    res = utils.read_in_json(os.path.join(working_path, f"{filename}.json"))
    os.remove(os.path.join(working_path, f"{filename}.json"))
    read_end = time.time()
    print(
        f'Reading and removing the json file took {"%.2f" % (read_end - subprocess_end)} seconds.')
    return res["data"]


def edit_wi_object(path, pgm_object, read_in_whitelists):
    return editing.edit_wi_object(path, pgm_object["structure"], read_in_whitelists)


# TODO: not needed -> in summary
def parse_object(pgm_object, wi_object, read_in_whitelists):
    # read in general structure
    return oty.parse_object(wi_object, pgm_object["structure"],
                            read_in_whitelists)



import copy

import src.utils as utils
import src.web_interface.wi_utils as wi_utils
import os


def parse_object(wi_object, key_yaml, read_in_whitelists):
    """
    This function parses a wi object back into a yaml
    :param key_yaml: the read in general structure
    :param wi_object: the filled wi object
    :return: result: a dictionary matching the metadata yaml structure
    """

    # define an empty dictionary to store the converted wi_object
    result = {}

    # TODO: needed? better solution?
    # save project id
    project_id = ''
    for elem in wi_object['project']['input_fields']:
        if elem['position'].split(':')[-1] == 'id' and elem['value'] \
                is not None:
            project_id = elem['value']

    # set parameters organism, sample_name and nom -> get filled during
    # conversion
    organism = ''
    sample_name = ''
    nom = 1
    global_count = 1
    local_count = 1
    double = []
    publication_records = wi_object['publication_records']
    # iterate over parts (from general structure to ensure order)
    for key in key_yaml:

        # make sure the key is present in the wi object
        if key in wi_object and key not in ['old_filenames', 'old_sample_names']:

            # parse every part into yaml format
            result[key], organism, sample_name, nom, global_count, local_count, double = parse_part(
                wi_object[key], key_yaml, copy.deepcopy(wi_object['all_factors']), project_id,
                organism, sample_name, nom, global_count, local_count, double, read_in_whitelists, publication_records)

    # remove keys with value None
    result = {k: v for k, v in result.items() if v is not None}

    sample_name_positions = get_sample_name_positions(result)
    for sample_pos in sample_name_positions:
        utils.fill_key(sample_pos, utils.create_sample_names(
            result, wi_object['old_sample_names'] if
            'old_sample_names' in wi_object else {}, sample_pos, read_in_whitelists=read_in_whitelists), result)

    filename_positions = get_filename_positions(result)
    for file_pos in filename_positions:
        utils.fill_key(file_pos, utils.create_filenames(
            result, double, file_pos, wi_object['old_filenames'] if
            'old_filenames' in wi_object else {}, read_in_whitelists=read_in_whitelists), result)

    return result


def get_filename_positions(metafile):
    positions = [x + ['filenames'] for x in utils.get_key_positions(metafile, 'technical_replicates', [], [])]
    return positions


def get_sample_name_positions(metafile):
    positions = [x + ['sample_name'] for x in
                 utils.get_key_positions(metafile, 'technical_replicates', [],
                                         [])]
    return positions


def split_name(elem, double, gene=True):
    new_name = []
    if '+' in elem:
        new_split = elem.split('+')
        for part in new_split:
            new_part, gene = split_name(part, double, gene=gene)
            if new_part != '':
                new_name.append(new_part)
        elem = '-'.join(new_name)

    if '#' in elem:
        remove = elem.split('#')[0]
        elem, gene = split_name(elem[len(f'{remove}#'):], double, gene=gene)

    if '.' in elem:
        if elem.lower() in [f'gn.{x.lower()}' for x in double]:
            gene = False
            elem = ''
        elif elem.lower().startswith('embl.') and gene is True:
            elem = ''
        else:
            elem = elem.split('.')[1]

    return elem, gene


def parse_part(wi_object, key_yaml, factors, project_id, organism, sample_name,
               nom, global_count, local_count, double, read_in_whitelists, publication_records):
    """
    This function parses a part of the wi object to create the yaml structure
    :param key_yaml: the read in general structure
    :param sample_name: the name of a sample build from condition name and
                        index of biological replicate
    :param wi_object: a part of the filled wi object
    :param factors: the selected experimental factors
    :param organism: the selected organism
    :param project_id: the project id
    :param nom: the number of measurements
    :return: val: the parsed part in yaml structure
             organism: the shortened version of the used organism
             sample_name: the name of the current sample
             nom: the number of measurements for the current sample
    """

    # initialize the converted value with None
    val = None

    # test if the object to parse is a dictionary
    if isinstance(wi_object, dict):

        # test if the values were stored in the 'list_value' key
        # (if the key takes a list or is of type single- or multi-autofill)
        if 'list_value' in wi_object and not (
                'input_type' in wi_object and wi_object['input_type'] in
                ['single_autofill', 'restricted_short_text']):

            # define an empty list to store the converted list values
            val = []
            is_cond = False
            # TODO: headers for list element (if gene is a list)
            # iterate over the list elements
            for i in range(len(wi_object['list_value'])):

                # test if the element is a dictionary
                if isinstance(wi_object['list_value'][i], dict):

                    # special case: condition
                    if wi_object['list_value'][i]['position'].split(
                            ':')[-1] == 'condition':

                        # define emtpy list to save parsed samples to
                        samples = []
                        local_count = 1
                        is_cond = True
                        # iterate over all samples
                        for sub_elem in \
                                wi_object['list_value'][i]['list_value']:

                            # convert samples
                            sample, organism, sample_name, nom, global_count, local_count, double = \
                                parse_list_part(sub_elem, key_yaml, factors,
                                                project_id, organism,
                                                sample_name, nom, global_count, local_count, double, read_in_whitelists, publication_records)

                            # remove empty keys
                            sample = {k: v for k, v in sample.items() if v is
                                      not None}

                            # add sample to the samples list
                            samples.append(sample)

                        # add dictionary with condition name and biological
                        # replicates (samples) to list parsed list
                        condition = (
                            {'condition_name': wi_object['list_value'][
                                i]['correct_value'],
                             'biological_replicates': {'count': len(samples)}})

                        # add samples to the condition dictionary if count > 0
                        if len(samples) > 0:
                            condition['biological_replicates']['samples'] = \
                                samples

                        # add condition to list of converted values
                        val.append(condition)

                # test if list element is a list
                elif isinstance(wi_object['list_value'][i], list):

                    # special case: experimental setting
                    if wi_object['position'].split(':')[-1] == \
                            'experimental_setting':

                        # call parse function using the experimental factors
                        # with the same index as the setting
                        c_val, organism, sample_name, nom, global_count, local_count, double = \
                            parse_list_part(wi_object['list_value'][i],
                                            key_yaml, factors[i], project_id,
                                            organism, sample_name, nom, global_count, local_count, double, read_in_whitelists, publication_records)

                        # add the converted value to the list
                        val.append(c_val)

                    # no special case
                    else:

                        # call parse function with all experimental factors
                        c_val, organism, sample_name, nom, global_count, local_count, double = \
                            parse_list_part(wi_object['list_value'][i],
                                            key_yaml, factors, project_id,
                                            organism, sample_name, nom, global_count, local_count, double, read_in_whitelists, publication_records)

                        # add the converted value to the list
                        val.append(c_val)

                # if list element is str/int/bool
                else:

                    new_val = copy.deepcopy(wi_object['list_value'][i])
                    parse_w_key = None
                    if 'whitelist_keys' in wi_object:
                        for w_key in wi_object['whitelist_keys']:
                            if new_val.endswith(f'({w_key})'):
                                new_val = new_val.rstrip(
                                    f'({w_key})').strip()
                                parse_w_key = w_key
                                break

                    if 'headers' in wi_object:
                        if isinstance(wi_object['headers'], dict):
                            if parse_w_key is not None:
                                if parse_w_key in wi_object['headers']:
                                    new_val = wi_utils.parse_headers(
                                        wi_object['headers'][parse_w_key],
                                        new_val, mode='dict')
                        else:
                            new_val = wi_utils.parse_headers(
                                wi_object['headers'],
                                new_val, mode='dict')
                    # add list element to list
                    val.append(new_val)

            # set the converted value to None if it is an empty list
            if len(val) == 0:
                val = None
            elif is_cond:
                val = sorted(val, key=lambda x: x['condition_name'])

        # the values are not saved as a list
        else:

            # wi object contains input fields
            if 'input_fields' in wi_object:

                # special case: technical replicates
                if wi_object['position'].split(':')[-1] == \
                        'technical_replicates':

                    #file_name = get_file_name(sample_name.removesuffix(f'_{sample_name.split("_")[-1]}'), double)
                    # TODO: comment
                    #t_sample_name = []
                    #t_file_name = []
                    count = [x['value'] for x in wi_object['input_fields'] if
                             x['position'].split(':')[-1] == 'count'][0]

                    #for c in range(count):
                    #    for m in range(nom):
                    #        t_sample_name.append(f'{project_id}_'
                    #                             f'{organism}_'
                    #                             f'{sample_name}'
                    #                             f'_t{"{:02d}".format(c + 1)}_'
                    #                             f'm{"{:02d}".format(m + 1)}')
                    #        t_file_name.append(f'{project_id}__'
                    #                           f'{global_count}__'
                    #                           f'{file_name}__'
                    #                           f'{local_count}')
                    #        local_count += 1
                    #        global_count += 1

                    val = {'count': count}

                # no special case
                else:

                    # call this function on the input fields
                    val, organism, sample_name, nom, global_count, local_count, double = \
                        parse_part(wi_object['input_fields'], key_yaml,
                                   factors, project_id, organism, sample_name,
                                   nom, global_count, local_count, double, read_in_whitelists, publication_records)

            # no input fields
            else:

                # set the value that should be converted (saved in list_value
                # if input type is 'single_autofill', else saved in value)
                convert_value = wi_object['list_value'][0] if \
                    wi_object['input_type'] in ['single_autofill', 'restricted_short_text'] and \
                    len(wi_object['list_value']) > 0 else wi_object['value']

                # test if value was filled
                if convert_value is not None:

                    # wi object contains whitelist keys
                    if 'whitelist_keys' in wi_object:
                        
                        # replace value with converted one
                        convert_value = wi_utils.parse_whitelist_keys(
                            wi_object['whitelist_keys'], convert_value,
                            wi_object['headers']
                            if 'headers' in wi_object else None)

                    # wi object contains headers but no whitelist keys
                    elif 'headers' in wi_object:

                        # replace the original value with the one split
                        # according to the header
                        convert_value = wi_utils.parse_headers(
                            wi_object['headers'], convert_value)

                    # value is of type value_unit
                    if wi_object['input_type'] == 'value_unit':

                        # save the value and unit as a dictionary
                        val = {'value': wi_object['value'],
                               'unit': wi_object['value_unit']}

                    # value is of type date
                    elif wi_object['input_type'] == 'date':

                        # convert the date to a string of format 'DD.MM.YYYY'
                        val = wi_utils.date_to_str(wi_object['value'])

                    # value was changed for display -> original value saved at
                    # key 'correct_value'
                    elif 'correct_value' in wi_object:

                        # special_case sample name
                        if wi_object['position'].split(':')[-1] == \
                                'sample_name':

                            # split and save index of sample from sample name
                            sample_count = int(
                                convert_value.split('_')[-1])

                            # reformat sample name with ending 'b<index>' for
                            # number of biological replicate
                            val = f'{wi_object["correct_value"]}_b'\
                                  f'{"{:02d}".format(sample_count)}'

                            # set the sample name to the value
                            sample_name = val

                        else:

                            # save correct value
                            val = wi_object['correct_value']

                    else:

                        if isinstance(convert_value, str):

                            info = list(utils.find_keys(
                                key_yaml,
                                wi_object['position'].split(':')[-1]))
                            if len(info) > 0 and 'input_type' in info[0] and \
                                    info[0]['input_type'] == 'bool':
                                if convert_value == 'True':
                                    val = True
                                else:
                                    val = False
                            else:
                                val = convert_value.replace('\\n', ' ').replace('\n', ' ').replace('\\', ' ')
                                val = ' '.join(val.split())
                        else:
                            # save value
                            val = convert_value

                        # set the number of replicates
                        if wi_object['position'].split(':')[-1] == \
                                'number_of_measurements':
                            nom = val

    # wi object is a list
    elif isinstance(wi_object, list):

        # call parse list function
        val, organism, sample_name, nom, global_count, local_count, double = parse_list_part(
            wi_object, key_yaml, factors, project_id, organism, sample_name,
            nom, global_count, local_count, double, read_in_whitelists, publication_records)

    return val, organism, sample_name, nom, global_count, local_count, double


def parse_list_part(wi_object, key_yaml, factors, project_id, organism,
                    sample_name, nom, global_count, local_count, double, read_in_whitelists, publication_records):
    """
    This function parses a part of the wi object of type list into the yaml
    structure
    :param sample_name: the name of a sample build from condition name and
                        index of biological replicate
    :param key_yaml: the read in general structure
    :param wi_object: the part of the wi object
    :param factors: the selected experimental factors
    :param organism: the selected organism
    :param project_id: the project id
    :param nom: the number of measurements
    :return: res: the parsed part in yaml structure
             organism: the shortened version of the used organism
             sample_name: the name of the current sample
             nom: the number of measurements for the current sample
    """

    # create dictionary to save parsed elements to
    res = {}

    # iterate over list
    for i in range(len(wi_object)):

        val = None

        # special case: organism
        if wi_object[i]['position'].split(':')[-1] == 'organism':

            # TODO: headers
            # save the organism name
            organism = wi_object[i]['value'].split(' ')[0]

            # read in the abbrev whitelist for organisms
            short = utils.get_whitelist(os.path.join(
                'abbrev', 'organism_name'),
                {'organism': organism}, whitelist_object=read_in_whitelists)['whitelist']

            # save the shortened version of the organism
            organism = short[organism]

        # special case: experimental factors
        if wi_object[i]['position'].split(':')[-1] == 'experimental_factors':

            # iterate over experimental factors
            for r in range(len(factors)):

                factors[r], double = parse_factor(factors[r], key_yaml, double)

            # set val to factors
            val = factors

        elif wi_object[i]['position'].split(':')[-1] == 'publication':
            publications = []
            for elem in wi_object[i]['list_value']:
                pubmed_id = None
                for sub_elem in elem:
                    if sub_elem['position'].split(':')[-1] == 'pubmed_id':
                        pubmed_id = sub_elem['value']
                        break
                if pubmed_id is not None:
                    if str(pubmed_id) not in publication_records:
                        publications.append({'pubmed_id': pubmed_id})
                    else:
                        pub_keys = list(utils.find_keys(key_yaml, 'publication'))[0]
                        publication = {}
                        for key in pub_keys['value']:
                            if key in publication_records[str(pubmed_id)]:
                                publication[key] = publication_records[str(pubmed_id)][key]
                        publications.append(publication)
            if len(publications) > 0:
                val = publications
        # no special case
        else:

            # call parse part function
            val, organism, sample_name, nom, global_count, local_count, double = parse_part(
                wi_object[i], key_yaml, factors, project_id, organism,
                sample_name, nom, global_count, local_count, double, read_in_whitelists, publication_records)

        # test if the value is not empty
        if val is not None and (
                (type(val) in [str, list, dict] and len(val) > 0) or
                type(val) not in [str, list, dict]):

            # overwrite the old value with the converted one
            res[wi_object[i]['position'].split(':')[-1]] = val

    return res if len(res) > 0 else None, organism, sample_name, nom, global_count, local_count, double


def parse_factor(factors, key_yaml, double):

    # nested factor
    if len(factors['values']) == 1 and isinstance(
            factors['values'][0], dict):

        if 'nested_infos' in factors:
            for key_ in factors['nested_infos']:
                if 'double' in factors['nested_infos'][key_]:
                    double = list(set(double + factors['nested_infos'][key_]['double']))
                

        # remove key 'multi'
        if 'multi' in factors['values'][0]:
            factors['values'][0].pop('multi')

        # remove keys with None as value
        factors['values'][0] = {
            k: v for k, v in factors['values'][0].items()
            if v is not None}

        # test if the key in the 'values' dictionary matches the
        # factor and overwrite the dictionary with its list values
        # -> this is the case if a factor can be a list and can
        #    therefor occur multiple times in a condition
        # -> key multi has to be added -> change 'values' from list
        #    to dict
        # -> e.g. factor tissue -> values {tissue: [...], multi: }
        if list(factors['values'][0].keys()) == \
                [factors['factor']]:
            factors['values'] = \
                factors['values'][0][factors['factor']]

        for i in range(len(factors['values'])):
            if isinstance(factors['values'][i], dict):
                remove_keys = []
                for key in factors['values'][i]:

                    whitelist_keys = None
                    headers = None

                    if 'nested_infos' in factors and key in factors['nested_infos']:
                        if 'whitelist_keys' in factors['nested_infos'][key]:
                            whitelist_keys = factors['nested_infos'][key][
                                'whitelist_keys']
                        if 'headers' in factors['nested_infos'][key]:
                            headers = factors['nested_infos'][key]['headers']
                        
                    if factors['values'][i][key] is None or (isinstance(factors['values'][i][key], list) and len(factors['values'][i][key]) == 0):
                        remove_keys.append(key)

                    if whitelist_keys is not None:

                        for j in range(len(factors['values'][i][key])):
                            factors['values'][i][key][j] = \
                                wi_utils.parse_whitelist_keys(
                                    whitelist_keys,
                                    factors['values'][i][key][j], headers,
                                    mode='dict')
                    elif headers is not None:
                        for j in range(len(factors['values'][i][key])):
                            factors['values'][i][key][j] = \
                                wi_utils.parse_headers(
                                    headers, factors['values'][i][key][j],
                                    mode='dict')

                    #if isinstance(factors['values'][i][key], dict):
                    #    if not 'factor' in factors['values'][i][key]:
                    #        factors['values'][i][key]['factor'] = key
                    #    factors['values'][i][key] = parse_factor(factors['values'][i][key], key_yaml)
                    #    factors['values'][i][key].pop('factor')
                for r_key in remove_keys:
                    factors['values'][i].pop(r_key)

    else:

        # fetch the properties of the experimental factor from the
        # general structure
        infos = list(utils.find_keys(
            key_yaml, factors['factor']))

        if 'double' in factors:
            double = set(double + factors['double'])

        # iterate over values of experimental factor
        for j in range(len(factors['values'])):

            # factor contains whitelist keys
            if 'whitelist_keys' in factors:

                # replace value with converted one
                factors['values'][j] = \
                    wi_utils.parse_whitelist_keys(
                        factors['whitelist_keys'],
                        factors['values'][j], factors['headers']
                        if 'headers' in factors else None)

            # factor contains headers but no whitelist keys
            elif 'headers' in factors:

                # replace the original value with the one split
                # according to the header
                factors['values'][j] = wi_utils.parse_headers(
                    factors['headers'], factors['values'][j])

            # factor of type value_unit
            elif len(infos) > 0 and 'special_case' in infos[0] \
                    and 'value_unit' in infos[0]['special_case']:

                factors['values'][j] = \
                    utils.split_value_unit(
                        factors['values'][j])

    # remove the keys 'header' and 'whitelist_keys'
    if 'whitelist_keys' in factors:
        factors.pop('whitelist_keys')
    if 'headers' in factors:
        factors.pop('headers')
    if 'nested_infos' in factors:
        factors.pop('nested_infos')
    if 'double' in factors:
        factors.pop('double')
    return factors, double


import src.utils as utils
import src.web_interface.wi_object_to_yaml as oty
import src.heatmap.create_heatmap as create_heatmap
from jinja2 import Template
import os
import plotly.graph_objects as go
import plotly.io as pio
import plotly as plt


def get_summary(wi_object, key_yaml, read_in_whitelists):
    """
    This function parses the wi object into a yaml structure and then parses
    the yaml to HTML to be output in the web interface. It also returns a list
    of filenames
    :param key_yaml: the read in general structure
    :param wi_object: the filled wi object
    :return: a dictionary containing the yaml structure as a dictionary and as
             HTML as well as the filenames as a string and in HTML
    """

    # parse wi_object to yaml
    yaml_object = oty.parse_object(wi_object, key_yaml, read_in_whitelists)
    
    # save the project_id from the yaml file
    if 'project' in yaml_object and 'id' in yaml_object['project']:
        project_id = yaml_object['project']['id']
    else:
        project_id = None

    # fetch all filenames from the yaml via a generator -> nested lists
    filename_nested = list(utils.find_list_key(
        yaml_object, 'technical_replicates:filenames'))

    # save filenames in html and string format
    filenames = get_html_filenames(filename_nested)

    # rewrite yaml to html
    
    template = Template(
        '''              
        {% if input.html %}
            {{ input.html }}
        {% else %}            
            <div style="overflow:auto; overflow-y:hidden; margin:0 auto; white-space:nowrap; padding-top:20">
                    {% if input.plot %}
                        {{ input.plot }}
                    {% endif %}
                    
                    {% if input.missing_samples %}
                        <i>Conditions without samples:</i>
                        {{ input.missing_samples }}
                    {% endif %}
            </div>
        {% endif %}
        '''
        )
    
    summary = {}
    setting_dict = {}
    for elem in ['project', 'experimental_setting', 'technical_details']:
        setting_dict[elem] = [] if elem == 'experimental_setting' else ''
        if elem in yaml_object:
            header =  elem.replace("_", " ").title()
            if elem == 'experimental_setting':
                plots = create_heatmap.get_heatmap(yaml_object, key_yaml, show_setting_id=False)
                plot_list = []
                for plot in plots:
                    add_plot = {'header': header}
                    if plot[1] is not None:
                        add_plot['plot'] = plot[1].to_html(full_html=False, include_plotlyjs='cdn') if plot[1] is not None else plot[1]
                    if plot[2] is not None:
                        add_plot['missing_samples'] = object_to_html(plot[2], 0, False)
                    plot_list.append({'title': plot[0], 'plot': template.render(input=add_plot)})
                setting_dict[elem] = plot_list
            else:
                setting_dict[elem] = template.render(input={'header': header, 'html': object_to_html(yaml_object[elem], 0, False)})

    summary['files'] = {'file_names': filenames,
                        'file_string': (project_id, '\n'.join(filenames)) if project_id is not None else None}
    summary['summary'] = setting_dict
    return summary


def get_html_filenames(filename_nest):
    """
    This function parses the filenames into HTML
    :param filename_nest: a nested list of filenames
    :return:
    html_filenames: the file names in HTML format
    filenames: the filenames as a list of strings
    """

    # define empty list to store filenames in string format
    filenames = []

    # iterate over nested filenames
    for file_list in filename_nest:

        # iterate over single filenames
        for filename in file_list:

            # add the filename to the html and to the list
            filenames.append(filename)

    return filenames


def object_to_html(yaml_object, depth, is_list):
    """
    This function parses the yaml structure into HTML
    :param yaml_object: a dictionary containing the yaml format
    :param depth: the depth of the indentation
    :param is_list: a boolean to state if a key contains a list
    :return: html_str: the yaml structure in HTML
    """

    # initialize html string
    html_str = ''

    # yaml is a dictionary
    if isinstance(yaml_object, dict):

        # iterate over keys in dictionary
        for key in yaml_object:

            # first key in a list -> bullet point
            if key == list(yaml_object.keys())[0] and is_list:

                # convert value of key to html
                input_text = object_to_html(yaml_object[key], depth + 1,
                                            is_list)

                # call function get_color to select the color of the key and
                # add key and html value to html string with a bullet point
                html_str = f'{html_str}<ul class="list-style-type-circle">' \
                           f'<li><p><font color={get_color(depth)}>{key}' \
                           f'</font>: {input_text}</p></li></ul>'

            # key without bullet point
            else:

                # convert value of key to html
                input_text = object_to_html(yaml_object[key], depth + 1,
                                            is_list)

                # call function get_color to select the color of the key and
                # add key and html value to html string
                html_str = f'{html_str}<ul class="list-style-none"><li><p>' \
                           f'<font color={get_color(depth)}>{key}</font>: ' \
                           f'{input_text}</p></li></ul>'

    # yaml is a list
    elif isinstance(yaml_object, list):

        # iterate over list elements
        for elem in yaml_object:

            # list element is single value
            if not isinstance(elem, list) and not isinstance(elem, dict):

                # add value to html string with a bullet point
                html_str = f'{html_str}<ul class="list-style-type-circle">' \
                           f'<li><p>{elem}</p></li></ul>'

            # list element is dict or list
            else:

                # call this function on list element and add it to html string
                html_str = f'{html_str}{object_to_html(elem, depth, True)}'

    # yaml is a single value
    else:

        # add value to html string
        html_str = f'{html_str}{yaml_object}'

    return html_str


# TODO: new color scheme -> 2 colors as parameter
def get_color(depth):
    """
    This function returns a color for the key in the HTML format depending on
    its indentation
    :param depth: the depth of indentation
    :return: color: the color in which the key should be colored
    """

    # color 1 if number of indentations is even
    if depth % 2 == 0:
        color = '26a69a'

    # color 2 if number of indentations is uneven
    else:
        color = '#d95965'

    return color



from src.input_functions import Input
from src.autogenerate import Autogenerate
from src import validate_yaml
from src import utils
from src.heatmap import create_heatmap
import os
from jinja2 import Template

template = Template(
        '''
        <h3>{{ header }}</h3>
                        
            {% for elem in plots %}

                {% if loop.index0 != 0 %}
                    <hr style="border-style: dotted;" />
                {% endif %}
                                
                <div style="woverflow:auto; overflow-y:hidden; margin:0 auto; white-space:nowrap; padding-top:5">
                    {{ elem.plot }}

                    {% if elem.missing_samples %}
                        <i>Conditions without samples:</i>
                        {{ elem.missing_samples }}
                    {% endif %}
                </div>
                                
            {% endfor %} 
                
        '''
        )

class Generate(Input):

    def generate(self):
        indent = 1
        for part in self.key_yaml:
            self.parse_lists(self.key_yaml[part], [part], indent,
                             self.result_dict)
            if part == 'experimental_setting':
                plot = create_heatmap.get_heatmap({part: self.result_dict[part]}, self.key_yaml)
                plots = [{'plot': elem[1], 'missing_samples': elem[2]}]
                for elem in plot:
                    if elem[1] is not None:
                        elem[1].show()
            else:
                print(self.get_summary(self.result_dict[part]))
        for elem in self.generate_end:
            func = getattr(Autogenerate, f'get_{elem[-1]}')
            fill_val = func(
                Autogenerate(self, elem))
            if fill_val is not None:
                self.fill_key(elem, fill_val, self.result_dict)

        # print validation report
        #print(self.get_validation(self.result_dict))
        print(self.get_summary(self.result_dict))
        # save information to a yaml file and print the filename
        print(
            f'File is saved to '
            f'{os.path.join(self.path, f"{self.project_id}{self.filename}.yaml")}'
            f'')
        utils.save_as_yaml(self.result_dict, os.path.join(
            self.path, f'{self.project_id}{self.filename}.yaml'))

    def print_sample_names(self):
        """
        This function creates a string out of all generated filenames that can
        be printed.
        """
        samples = list(
            utils.find_list_key(self.result_dict,
                                'technical_replicates:sample_name'))
        print(f'{"".center(self.size, "-")}\n'
              f'{"SAMPLE NAMES".center(self.size, " ")}\n'
              f'{"".center(self.size, "-")}\n')
        sample_names = ''
        for elem in samples:
            for name in elem:
                sample_names += f'- {name}\n'
        print(sample_names)
        save = self.parse_list_choose_one(
            ['True ', 'False '],
            'Do you want to save the sample names into a file?')
        if save:
            text_file = open(os.path.join(
                self.path, f'{self.project_id}_samples.txt'), 'w')
            text_file.write(sample_names)
            text_file.close()
            print(
                f'The sample names have been saved to file '
                f'\'{self.path}/{self.project_id}_samples.txt\'.')

    def get_summary(self, result):
        summary = ''
        summary += f'{"".center(self.size, "=")}\n' \
                   f'{"SUMMARY".center(self.size, " ")}\n' \
                   f'{"".center(self.size, "=")}\n'
        summary += self.print_summary(result, 1, False)
        summary += f'\n\n'
        summary += f'{"".center(self.size, "=")}\n'
        return summary

    def get_validation(self, result):
        validation_reports = {'all_files': 1,
                              'corrupt_files': {'count': 0, 'report': []},
                              'error_count': 0, 'warning_count': 0}
        file_reports = {'file': result, 'error': None, 'warning': None}
        report = ''
        report += f'{"FILE VALIDATION".center(self.size, " ")}\n' \
                  f'{"".center(self.size, "-")}\n'
        valid, missing_mandatory_keys, invalid_keys, \
            invalid_entries, invalid_values, logical_warn = \
            validate_yaml.validate_file(result, self.key_yaml, self.filename)
        if not valid:
            validation_reports['corrupt_files']['count'] = 1
            validation_reports['error_count'] += (
                    len(missing_mandatory_keys) + len(invalid_keys) + len(
                        invalid_entries) + len(invalid_values))
            file_reports['error'] = (
                missing_mandatory_keys, invalid_keys, invalid_entries,
                invalid_values)
        if len(logical_warn) > 0:
            validation_reports['corrupt_files']['count'] = 1
            validation_reports['warning_count'] += len(logical_warn)
            file_reports['warning'] = logical_warn
        validation_reports['corrupt_files']['report'].append(file_reports)

        report += f'Found {validation_reports["error_count"]} errors and ' \
                  f'{validation_reports["warning_count"]} warnings.\n'

        if validation_reports['corrupt_files']['count'] > 0:
            rep = ''
            for _report in validation_reports['corrupt_files']['report']:
                rep += f'{"".center(self.size, "_")}\n\n'
                rep += validate_yaml.print_full_report(_report['file'],
                                                       _report['error'],
                                                       _report['warning'],
                                                       self.size)
            rep += f'{"".center(self.size, "_")}\n\n'
            report += rep
        return report

    def print_summary(self, result, depth, is_list):
        """
        This function parses the dictionary into a string with the same
        structure as it will be saved to the yaml file
        :param result: the filled dictionary
        :param depth: an integer that specifies the depth of indentation
        :param is_list: a bool that states if a key contains a list
        :return: summary: a string that contains all entered information
        """
        summary = ''
        if isinstance(result, dict):
            for key in result:
                printed_summary = self.print_summary(result[key],
                                                     depth + 1, is_list)
                if key == list(result.keys())[0] and is_list:
                    summary = f'{summary}\n{"    " * (depth - 1)}{"  - "}' \
                              f'{key}: {printed_summary}'
                else:
                    summary = f'{summary}\n{"    " * depth}{key}: ' \
                              f'{printed_summary}'
        elif isinstance(result, list):
            for elem in result:
                if not isinstance(elem, list) and not isinstance(elem, dict):
                    summary = f'{summary}\n{"    " * (depth - 1)}{"  - "}' \
                              f'{elem}'
                else:
                    summary = f'{summary}' \
                              f'{self.print_summary(elem, depth, True)}'
        else:
            summary = f'{summary}{result}'
        return summary

    def parse_lists(self, structure, position, indent, return_dict,
                    is_factor=False):

        if isinstance(structure['value'], dict):
            elem_index = 0
            redo = True

            while redo:
                self.print_header(structure['display_name'], indent)
                print(utils.print_desc(structure['desc'], size=self.size),
                      '\n')
                is_list = False
                if 'special_case' in structure and 'merge' in structure['special_case']:
                    self.fill_key(position + [elem_index] if structure['list'] or is_factor else position, self.parse_input_value(
                            position[-1], structure['value'][structure[
                                'special_case']['merge']]), return_dict)
                    if is_factor:
                        is_list = True
                elif 'special_case' in structure and 'value_unit' in structure['special_case']:
                    if structure['list'] or is_factor:
                        self.fill_key(
                                position,
                                self.get_list_value_unit(position[-1], structure),
                                return_dict)
                    else:
                        self.fill_key(
                                position,
                                self.get_value_unit(structure), return_dict)
                else:
                    self.input_keys(
                        structure['value'],
                        position + [elem_index] if structure['list'] or is_factor else position,
                        indent, return_dict, is_factor=is_factor)
                self.print_headline(indent)
                if (structure['list'] and not is_factor) or (is_list and is_factor):
                    redo = self.parse_list_choose_one(
                            ['True ', 'False '],
                            f'\nDo you want to add another '
                            f'{structure["display_name"]}?')
                    elem_index += 1
                else:
                    redo = False

        else:
            self.fill_key(
                position, self.get_input_list(structure, position[-1]) if
                structure['list'] or is_factor else self.parse_input_value(
                    position[-1], structure),
                return_dict)

    def print_headline(self, indent):
        if indent > 1:
            delim = '-'
        else:
            delim = '_'
        print(f'\n{"".center(self.size, delim)}\n')

    def print_header(self, key, indent):
        if indent > 1:
            delim = '-'
            new_line = ''
        else:
            delim = '_'
            new_line = '\n'
        print(f'\n'
              f'{"".center(self.size, delim)}{new_line}\n'
              f'{f"{key}".center(self.size, " ")}\n'
              f'{"".center(self.size, delim)}\n')

    def input_keys(self, structure, position, indent, return_dict,
                   is_factor=False):
        optionals = []
        desc = []
        for key in structure:
            if structure[key]['mandatory'] or \
                    ('special_case' in structure[key] and ('generated' in
                     structure[key]['special_case'] or 'factor' in structure[key]['special_case'])):
                if 'special_case' in structure[key]:
                    if 'factor' in structure[key]['special_case'] and \
                            structure[key]['special_case']['factor']:
                        self.fill_key(position + [key],
                                      structure[key]['value'], return_dict)
                    elif 'generated' in structure[key]['special_case']:
                        if structure[key]['special_case']['generated'] == \
                                'now':
                            func = getattr(Autogenerate, f'get_{key}')
                            fill_val = func(Autogenerate(self, position + [key]))
                            if fill_val is not None:
                                self.fill_key(position + [key], fill_val,
                                              return_dict)
                                if not isinstance(structure[key]['value'], dict):
                                    print(f'\n---\n{structure[key]["desc"]}\n')
                                    print(f'{key}: {utils.find_position(self.result_dict, position + [key])}')
                        elif structure[key]['special_case']['generated'] == \
                                'end':
                            self.generate_end.append(position + [key])
                        elif structure[key]['special_case']['generated'] == \
                                'fill':
                            self.fill_key(position + [key],
                                          structure[key]['value'], return_dict)
                            optionals.append(key)
                            desc.append(structure[key]['desc'])
                    elif 'value_unit' in structure[key]['special_case']:
                        if 'list' in structure[key]:
                            self.fill_key(
                                position + [key],
                                self.get_list_value_unit(key, structure),
                                return_dict)
                        else:
                            self.fill_key(
                                position + [key],
                                self.get_value_unit(structure), return_dict)
                    elif 'merge' in structure[key]['special_case']:
                        self.fill_key(
                            position + [key], self.parse_input_value(
                                key,
                                structure[key]['value'][structure[key][
                                    'special_case']['merge']]),
                            return_dict)
                    else:
                        self.parse_lists(structure[key], position + [key],
                                         indent + 1, return_dict,
                                         is_factor=is_factor)
                else:
                    self.parse_lists(structure[key], position + [key],
                                     indent+1, return_dict,
                                     is_factor=is_factor)
            else:
                if not ('special_case' in structure[key] and 'factor' in
                        structure[key]['special_case'] and
                        structure[key]['special_case']['factor']) or \
                        structure[key]['list']:
                    optionals.append(key)
                    desc.append(structure[key]['desc'])
        if len(optionals) > 0 and not self.mandatory_only:
            print(
                f'\nDo you want to add any of the following optional keys?'
                f' (1,...,{len(optionals)} or n)\n')
            self.print_option_list(optionals, desc)
            options = self.parse_input_list(optionals, True)
            if options:
                for option in options:
                    if 'special_case' in structure[option] and 'generated' in \
                            structure[option]['special_case'] and \
                            structure[option]['special_case']['generated'] != \
                            'fill':
                        if structure[option]['special_case']['generated'] == \
                                'now':
                            func = getattr(Autogenerate, f'get_{option}')
                            fill_val = func(
                                Autogenerate(self, position + [option]))
                            if fill_val is not None:
                                self.fill_key(position + [option], fill_val,
                                              return_dict)
                        elif structure[option]['special_case']['generated'] \
                                == 'end':
                            self.generate_end.append(position + [option])
                    else:
                        self.parse_lists(structure[option],
                                         position + [option], indent+1,
                                         return_dict, is_factor=is_factor)

    def fill_key(self, position, value, fill_dict):
        if len(position) > 0:
            if len(position) == 1:
                fill_dict[position[0]] = value
            else:
                if type(position[1]) == str:
                    if type(position[0]) == str and position[0] not in \
                            fill_dict:
                        fill_dict[position[0]] = {}
                else:
                    if position[0] not in fill_dict:
                        fill_dict[position[0]] = []
                    if len(fill_dict[position[0]]) < position[1]+1:
                        for i in range(len(fill_dict[position[0]]),
                                       position[1]+1):
                            fill_dict[position[0]].append({})
                self.fill_key(position[1:], value, fill_dict[position[0]])
        else:
            print('NO POSITION')



import os.path
import datetime
from src import utils


# This script includes functions for the validation of metadata yaml files

# ---------------------------------VALIDATION-----------------------------------

generated = ['condition_name', 'sample_name']
factor = None


def validate_file(metafile, key_yaml, filename, logical_validation=True, yaml=None, generated=True, whitelist_path=None, only_mandatory = False):
    """
    In this function all functions for the validation of a metadata file are
    called. The validation is based on the data in the file 'keys.yaml'. It is
    tested if all mandatory keys are included, if the included keys are valid
    and if the entered values correspond to the whitelist.
    :param metafile: the read in metadata yaml file
    :return:
    valid: bool, true if the file is valid, false if it is invalid
    missing_mandatory_keys: a list containing the missing mandatory keys
    invalid_keys: a list containing the invalid keys
    invalid_entries: a list containing the invalid entries -> (key, [values])
    """
    logical_warn = []
    valid = True

    if not only_mandatory:
        invalid_keys, invalid_entries, invalid_value = \
            new_test(metafile, key_yaml, [], '', [], [], [], None, [], None, metafile, key_yaml, whitelist_path=whitelist_path, filename=filename)
    else:
        invalid_keys = []
        invalid_entries = []
        invalid_value = []
    missing_mandatory_keys = test_for_mandatory(metafile, key_yaml,
                                                [x.split(':')[-1] for x in
                                                 invalid_keys], generated)
    if len(missing_mandatory_keys) > 0 or len(invalid_keys) > 0 or len(
            invalid_entries) > 0 or len(invalid_value) > 0:
        valid = False
    if logical_validation:
        logical_warn = validate_logic(metafile, filename)
    return valid, missing_mandatory_keys, invalid_keys, invalid_entries, \
        invalid_value, logical_warn


# -----------------------------------REPORT-------------------------------------


def print_full_report(metafile, errors, warnings, size=80):
    report = ''
    try:
        input_id = metafile['project']['id']
    except KeyError:
        input_id = 'missing'
    try:
        path = metafile['path']
    except KeyError:
        path = 'missing'
    report += f'{"VALIDATION REPORT".center(size, "-")}\n' \
              f'Project ID: {input_id}\n' \
              f'Path: {path}\n\n' \

    if errors is not None:
        report += f'{print_validation_report(errors[0], errors[1], errors[2], errors[3], size)}\n'
    if warnings is not None:
        report += f'{print_warning(warnings,size)}\n'
    return report


def print_validation_report(missing_mandatory_keys, invalid_keys,
                            invalid_values, invalid_value, size=80):
    """
    This function outputs a report on invalid files. The report contains the ID
     of the project, the path to the file, as well as the missing mandatory
     keys, invalid keys and invalid entries.
    :param invalid_value: a list containing invalid values
    :param metafile: the metafile that is validated
    :param missing_mandatory_keys: a list containing all missing mandatory keys
    :param invalid_keys: a list containing all invalid keys
    :param invalid_values: a list containing all invalid values
    """
    invalid_entries = '\n- '.join(invalid_keys)
    missing = '\n- '.join(missing_mandatory_keys)
    whitelist_values = []
    for v in invalid_values:
        key = ':'.join(v.split(':')[:-1])
        entry = v.split(':')[-1]
        whitelist_values.append(entry + ' in ' + key + '\n')
    value = []
    for v in invalid_value:
        value.append(f'{v[0]}: {v[1]} -> {v[2]}')
    report = ''
    report += f'{"ERROR".center(size, "-")}\n\n'
    if len(invalid_keys) > 0:
        report += f'The following keys were invalid:\n'\
                  f'- {invalid_entries}\n'
    if len(missing_mandatory_keys) > 0:
        report += f'The following mandatory keys were missing:\n' \
                  f'- {missing}\n'
    if len(invalid_values) > 0:
        report += f'The following values do not match the whitelist:\n' \
                  f'- {"- ".join(whitelist_values)}\n'
    if len(invalid_value) > 0:
        report += f'The following values are invalid:\n' \
                  f'- {"- ".join(value)}\n'
    return report


def print_warning(logical_warn, size=80):
    """
    This function prints a warning message.
    :param metafile: the metafile that contains the warning
    :param pool_warn: a list of warnings concerning pooled and donor_count
    :param ref_genome_warn: a list of warnings concerning the reference_genome
    """

    report = ''
    report += f'{"WARNING".center(size, "-")}\n\n'
    if len(logical_warn) > 0:
        for elem in logical_warn:
            report += f'- {elem[0]}:\n{elem[1]}\n'
    #if len(pool_warn) > 0:
    #    for elem in pool_warn:
    #        print(f'- Sample \'{elem[0]}\':\n{elem[1]}')
    #if len(ref_genome_warn) > 0:
    #    for elem in ref_genome_warn:
    #        print(f'- Run from {elem[0]}:\n{elem[1]}')
    return report

# --------------------------------UTILITIES------------------------------------

def new_test(metafile, key_yaml, sub_lists, key_name, invalid_keys,
             invalid_entry, invalid_value, input_type, is_factor,
             local_factor, full_metadata, full_yaml, whitelist_path=None, filename='_metadata'):
    """
    This function test if all keys in the metadata file are valid.
    :param metafile: the metadata file
    :param key_yaml: the read in keys.yaml
    :param sub_lists: a list to save all items within a key if it has a list as
                      value
    :param key_name: the name of the key that is tested
    :param invalid_keys: a list of all invalid keys
    :param invalid_entry: a list of all invalid entries
    :param invalid_value: a list of all invalid values
    :param input_type: the input type that is expected for the value
    :param is_factor: a bool to state if the key is an experimental factor
    :param local_factor: a parameter to save the current experimental factor
    :return:
    invalid_keys: a list containing the invalid keys
    invalid_entries: a list containing the invalid entries
    invalid_value: a list containing the invalid values
    """
    if isinstance(metafile, dict) and not (
            'value' in metafile and 'unit' in metafile):
        for key in metafile:
            if not key_yaml and key_name.split(':')[-1] in is_factor or (key_name.split(':')[-1] == 'values' and local_factor is not None):
                new_yaml1 = full_yaml
                if key_name.split(':')[-1] in is_factor:
                    new_yaml = list(utils.find_keys(new_yaml1, key_name.split(':')[-1]))
                else:
                    new_yaml = list(utils.find_keys(new_yaml1, local_factor))
                if len(new_yaml) > 0:
                    if 'whitelist' in new_yaml[0] and new_yaml[0]['whitelist']:
                        if key_name.split(':')[-1] in is_factor:
                            w = utils.get_whitelist(key_name.split(':')[-1],
                                                    full_metadata, whitelist_path=whitelist_path)
                        else:
                            w = utils.get_whitelist(local_factor, full_metadata, whitelist_path=whitelist_path)
                        if w and 'headers' in w:
                            if isinstance(w['headers'], dict):
                                if 'whitelist_keys' in w:
                                    headers = []
                                    for w_k in w['whitelist_keys']:
                                        if w_k in w['headers']:
                                            headers += w['headers'][w_k].split(' ')
                            else:
                                headers = w['headers'].split(' ')
                            new_yaml[0]['value'] = headers
                    #TODO: enrichment
                    if new_yaml[0]['value'] is not None:
                        if key not in new_yaml[0]['value']:
                            if key_name.split(':')[-1] == 'enrichment_type' and key in ('gene_name', 'ensembl_id'):
                                pass
                            else:
                                invalid_keys.append(f'{key_name}:{key}')
                        elif isinstance(metafile[key], list) != new_yaml[0]['list']:
                            if (key_name.split(':')[-1] == 'enrichment_type' and key in ('gene_name', 'ensembl_id')) or (key_name.split(':')[-1] == 'values' and key in ['enrichment_type', 'modification']):
                                pass
                            else:
                                invalid_keys.append(f'{key_name}:{key}')
                else:
                    if key_name.split(':')[-1] == 'enrichment_type' and key in ('gene_name', 'ensembl_id'):
                        pass
                    else:
                        invalid_keys.append(f'{key_name}:{key}')
            elif not key_yaml or (key_yaml and key not in key_yaml):
                if key_name.split(':')[-1] == 'enrichment_type' and key in ('gene_name', 'ensembl_id'):
                    pass
                else:
                    invalid_keys.append(f'{key_name}:{key}')
            else:
                if isinstance(key_yaml, dict) and key in key_yaml and isinstance(key_yaml[key], dict):
                    if key == 'factor':
                        global factor
                        factor = metafile[key]
                        local_factor = metafile[key]
                        is_factor.append(metafile[key])
                    input_type = None
                    if key == 'values' and factor is not None:
                        node = list(utils.find_keys(key_yaml, factor))
                        if len(node) > 0:
                            if 'input_type' in node:
                                input_type = node['input_type']
                    elif 'list' in key_yaml[key] and isinstance(metafile, dict) and key in metafile and isinstance(metafile[key], list) != key_yaml[key]['list']:
                        if key_name.split(':')[-1] == 'enrichment_type' and key in ('gene_name', 'ensembl_id'):
                            pass
                        else:
                            invalid_keys.append(f'{key_name}:{key}')
                    elif 'input_type' in key_yaml[key]:
                        input_type = key_yaml[key]['input_type']
                    res_keys, res_entries, res_values = new_test(
                        metafile[key], key_yaml[key]['value'], sub_lists,
                        f'{key_name}:{key}' if key_name != '' else key,
                        invalid_keys, invalid_entry, invalid_value, input_type,
                        is_factor, local_factor, full_metadata, full_yaml, whitelist_path=whitelist_path, filename=filename)
                    invalid_keys = res_keys
    elif isinstance(metafile, list):
        for item in metafile:
            sub_lists.append(item)
            res_keys, res_entries, res_values = new_test(
                item, key_yaml, sub_lists, key_name, invalid_keys,
                invalid_entry, invalid_value, input_type, is_factor,
                local_factor, full_metadata, full_yaml, whitelist_path=whitelist_path, filename=filename)
            invalid_keys = res_keys
            sub_lists = sub_lists[:-1]
    else:
        #TODO: Value unitb + check for whitelist but fast
        #has_whitelist = list(utils.find_keys(utils.read_in_yaml(key_yaml_path), key_name.split(':')[-1]))
        #if len(has_whitelist) > 0 and 'whitelist' in has_whitelist[0]:
            #has_whitelist = True
        #else:
            #has_whitelist = False
        invalid = new_test_for_whitelist(key_name.split(':')[-1], metafile,
                                         sub_lists, whitelist_path=whitelist_path)
        if invalid:
            invalid_entry.append(f'{key_name}:{metafile}')

        inv_value, message = validate_value(metafile, input_type,
                                            key_name.split(':')[-1], filename=filename)

        if not inv_value:
            invalid_value.append((key_name, metafile, message))

    return invalid_keys, invalid_entry, invalid_value


def new_test_for_whitelist(entry_key, entry_value, sublists, whitelist_path=None):
    """
    This function tests if the value of a key matches the whitelist.
    :param entry_key: the key that is tested
    :param entry_value: the value that has to match the whitelist
    :param sublists: a list to save all items within a key if it has a list as
                      value
    :return: True if the entry does not match the whitelist else False
    """
    if entry_value == None:
        entry_value = 'None'
    if whitelist_path == None:
        whitelist_path = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), '..',
            'FRED_whitelists')

    whitelist = utils.read_whitelist(entry_key, whitelist_path=whitelist_path)
    if whitelist:
        whitelist = parse_whitelist(whitelist, sublists, whitelist_path)
    if whitelist:
        if entry_value not in whitelist:
            return True
    return False


def parse_whitelist(whitelist, sublists, whitelist_path):
    if whitelist['whitelist_type'] == 'plain':
        whitelist = whitelist['whitelist']
    elif whitelist['whitelist_type'] == 'depend':
        whitelist_key = whitelist['ident_key']
        value = None
        for i in reversed(range(len(sublists))):
            value = list(utils.find_keys(sublists[i], whitelist_key))
            if len(value) > 0:
                if len(value) == 1:
                    break
                else:
                    print("ERROR: multiple values")
                    break
        if value and value[0] in whitelist:
            whitelist = whitelist[value[0]]
        else:
            whitelist = utils.read_whitelist(value[0], whitelist_path=whitelist_path)
            if whitelist and whitelist['whitelist_type'] == 'plain':
                whitelist = whitelist['whitelist']
    elif whitelist['whitelist_type'] == 'group':
        whitelist = utils.read_grouped_whitelist(whitelist, {})
        if 'whitelist_type' in whitelist and whitelist['whitelist_type'] == 'plain_group':
            if whitelist['whitelist_keys']:
                for key in whitelist['whitelist_keys']:
                    whitelist['whitelist'] = [x.replace(f'({key})', '').strip() for x in whitelist['whitelist']]
        if 'whitelist' in whitelist:
            if not isinstance(whitelist['whitelist'], list):
                plain_whitelist = []
                for key in whitelist['whitelist']:
                    if isinstance(whitelist['whitelist'][key], list):
                        plain_whitelist += whitelist['whitelist'][key]
                    elif os.path.isfile(os.path.join(whitelist_path, 'whitelists', whitelist['whitelist'][key])):
                        sub_whitelist = utils.read_whitelist(whitelist['whitelist'][key], whitelist_path=whitelist_path)
                        parsed_sub_whitelist = parse_whitelist(sub_whitelist, sublists, whitelist_path)
                        plain_whitelist += parsed_sub_whitelist
                whitelist = plain_whitelist
            else:
                whitelist = whitelist['whitelist']
    if whitelist and not isinstance(whitelist, list) and not isinstance(
            whitelist, dict) \
            and os.path.isfile(os.path.join(
            whitelist_path, 'whitelists', whitelist)):
        whitelist = utils.read_whitelist(whitelist, whitelist_path=whitelist_path)
        if whitelist:
            whitelist = parse_whitelist(whitelist, sublists, whitelist_path)
    return whitelist

def test_for_mandatory(metafile, key_yaml, invalid_keys, generated):
    """
    This function calls a function to get the missing mandatory keys for every 
    part of the metadata object.
    :param metafile: the read in metadata yaml file
    :param key_yaml: the read in structure file 'keys.yaml'
    :param invalid_keys: a list of keys that are invalid and should be ignored
    :return: missing_keys: a list containing the missing mandatory keys
    """
    missing_keys = []
    for key in key_yaml:
        missing_keys += get_missing_keys(key_yaml[key], metafile,
                                         invalid_keys, key, [], len(metafile[key]) if key in metafile and key_yaml[key]['list'] else 1, generated)
    return missing_keys


def get_missing_keys(node, metafile, invalid_keys, pre, missing, list_len, generated):
    """
    This function tests if all mandatory keys from the structure file
    'keys.yaml' are present in the metadata file.
    :param node: a key within the read in structure file 'keys.yaml'
    :param metafile: the read in metadata file
    :param invalid_keys: a list containing invalid keys that should be ignored
    :param pre: a string to save and chain keys in order to save their position
    :param missing: a list to save the missing mandatory keys
    """

    metafile = find_key(metafile, pre)

    if len(metafile) == 0:
        if node['mandatory'] and (('special_case' in node and 'generated' in node['special_case'] and node['special_case']['generated'] == 'end' and generated) or not ('special_case' in node and 'generated' in node['special_case'] and node['special_case']['generated'] == 'end')):
            missing.append(pre)
    else:
         if pre.split(':')[-1] not in invalid_keys:
            if isinstance(node['value'], dict) and not set(['mandatory', 'list',
                                                    'desc', 'display_name',
                                                    'value']) <= \
            set(node['value'].keys()):
                if isinstance(metafile[0], list):
                    for elem in metafile[0]:
                        for key in node['value']:
                            missing = get_missing_keys(node['value'][key], elem,
                                           invalid_keys, pre + ':' + key,
                                           missing, len(metafile[0]) if node['list'] else 1, generated)
                else:
                    for key in node['value']:
                        missing = get_missing_keys(node['value'][key], metafile[0],
                                                   invalid_keys,
                                                   pre + ':' + key,
                                                   missing,
                                                   len(metafile[0]) if node[
                                                       'list'] else 1, generated)
    return list(set(missing))


def find_key(metafile, key):
    """
    This function searches for a key of the keys.yaml in the metafile.
    :param metafile: the read in metadata file
    :param key: a string of chained keys (key1:key2...)
    :param is_list: bool, true if the instance in the structure is a list
    :param values: the default entry for the key within the structure file
    :param invalid_keys: a list containing invalid keys that should be ignored
    :return: missing_keys: a list containing the missing mandatory keys
    """
    for k in key.split(':'):
        new_metafile = list(utils.find_keys(metafile, k))
    return new_metafile


def validate_value(input_value, value_type, key, filename='_metadata'):
    """
    This function tests if an entered value matches its type and contains
    invalid characters.
    :param input_value: the value to be valiated
    :param value_type: the type of which the value should be
    :param key: the key that contains the value
    :return:
    valid: a boolean that states if the value is valid
    message: a string that contains information about the error if tha value
             is invalid
    """
    valid = True
    message = None
    if input_value is not None:
        if value_type == 'bool':
            if input_value not in [True, False]:
                valid = False
                message = 'The value has to be of type bool (True or False).'
        elif value_type == 'number':
            if not isinstance(input_value, int):
                valid = False
                message = 'The value has to be an integer.'
        elif value_type == 'date':
            try:
                if filename == '_metadata':
                    input_date = input_value.split('.')
                    date_message = f'Input must be of type \'DD.MM.YYYY\'.'
                elif filename in ['_mamplan', 'mamplan']:
                    input_date = input_value.split('/')
                    date_message = f'Input must be of type \'DD/MM/YYYY\' or \'DD/MM/YY\'.'
                else:
                    input_date = input_value
                    date_message = f'Invalid date.'
                if len(input_date) != 3 or len(input_date[0]) != 2 or len(
                        input_date[1]) != 2 or not ((filename in ['_mamplan', 'mamplan'] and len(input_date[2]) == 2) or (len(input_date[2]) == 4 or len(input_date[2]) == 2)):
                    raise SyntaxError
                input_value = datetime.date(int(input_date[2]),
                                            int(input_date[1]),
                                            int(input_date[0]))
            except (IndexError, ValueError, SyntaxError) as e:
                valid = False
                message = date_message
        elif type(input_value) == str and ('\"' in input_value or '{' in input_value or '}' in
                 input_value or '|' in input_value) and key not in generated:
            if filename not in ['_mamplan', 'mamplan']:
                valid = False
                message = 'The value contains an invalid character ' \
                          '(\", {, } or |).'
    return valid, message


def validate_logic(metafile, filename='_metadata'):
    """
    This functions tests the logic of the input data.
    :param metafile: the metafile to be validated
    :return:
    pool_warn: a list containing warnings about the donor_count and pooled
    ref_genome_warn: a list containing warnings about the reference genome
    """
    logical_warn = []

    if filename == '_metadata':
        techniques = list(utils.find_keys(metafile, 'setting'))
        setting_ids = list(utils.find_keys(metafile, 'setting_id'))
        warning, warn_message = validate_techniques(setting_ids, techniques)
        if warning:
            logical_warn.append((f'Invalid techniques:', warn_message))
        invalid_file, invalid_sample, file_message, sample_message = validate_filenames(metafile)
        if invalid_file:
            logical_warn.append((f'Invalid number of filenames:', file_message))
        if invalid_sample:
            logical_warn.append((f'Invalid number of sample names:', sample_message))
        samples = list(utils.find_keys(metafile, 'samples'))
        for cond in samples:
            for sample in cond:
                if isinstance(sample, dict) and 'pooled' in sample and 'donor_count' in sample:
                    warning, warn_message = validate_donor_count(sample['pooled'],
                                                                 sample['donor_count'])
                    if warning:
                        logical_warn.append((f'Sample \'{sample["sample_name"]}\'', warn_message))
        organisms = list(utils.find_keys(metafile, 'organism_name'))
        runs = list(utils.find_keys(metafile, 'runs'))
        if len(runs) > 0:
            for run in runs[0]:
                if 'reference_genome' in run:
                    warning, warn_message = validate_reference_genome(
                        organisms, run['reference_genome'])
                    if warning:
                        logical_warn.append((f'Run from {run["date"]}', warn_message))
    elif filename in ['_mamplan', 'mamplan']:
        if 'tags' in metafile and 'organization' in metafile['tags'] and metafile['tags']['organization'] is not None:
            if 'public' in metafile['tags']['organization']:
                if 'pubmedid' not in metafile['tags'] or metafile['tags']['pubmedid'] is None:
                    logical_warn.append(('tags:pubmedid', 'The pubmed ID is missing for this public project'))
                if 'citation' not in metafile['tags'] or metafile['tags']['citation'] is None:
                    logical_warn.append(('tags:citation', 'The citation is missing for this public project'))
        if 'project' in metafile and 'id' in metafile['project'] and metafile['project']['id'] is not None and metafile['project']['id'] != metafile['project']['id'].lower():
            logical_warn.append(('project:id', 'The ID should be lowercase'))
    return logical_warn


def validate_filenames(metafile):
    invalid_file = False
    invalid_sample = False
    file_message = ''
    sample_message = ''
    technique = list(utils.find_keys(metafile, 'techniques'))
    if len(technique) > 0:
        techniques = {}
        for elem in technique[0]:
            if 'setting' in elem:
                techniques[elem['setting']] = elem['technique']
        settings = list(utils.find_keys(metafile, 'experimental_setting'))
        if len(settings) > 0:
            for setting in settings[0]:
                if 'setting_id' in setting and setting['setting_id'] in techniques:
                    used_techniques = techniques[setting['setting_id']]
                    samples = list(utils.find_keys(setting, 'samples'))
                    if len(samples) > 0:
                        for sample in samples[0]:
                            sample_name = sample['sample_name'] if 'sample_name' in sample else ''
                            m = sample['number_of_measurements'] if 'number_of_measurements' in sample else 0
                            t = sample['technical_replicates']['count'] if 'technical_replicates' in sample and 'count' in sample['technical_replicates'] else 0
                            tech = len(used_techniques)
                            file_count = m * t * tech
                            if file_count > 0:
                                if 'technical_replicates' in sample:
                                    if 'sample_name' in sample['technical_replicates']:
                                        if len(sample['technical_replicates']['sample_name']) != file_count:
                                            invalid_sample = True
                                            sample_message += f'Number of sample names ({len(sample["technical_replicates"]["sample_name"])}) for sample \'{sample_name}\' does not match expected number of sample names ({file_count})\n'
                                    if 'filenames' in sample['technical_replicates']:
                                        if len(sample['technical_replicates']['filenames']) != file_count:
                                            invalid_file = True
                                            file_message += f'Number of filenames ({len(sample["technical_replicates"]["sample_name"])}) for sample \'{sample_name}\' does not match expected number of filenames ({file_count})\n'
    return invalid_file, invalid_sample, file_message, sample_message


def validate_techniques(setting_ids, techniques):
    invalid = False
    message = None
    missing_technique = list(set(setting_ids)-set(techniques))
    unknown_techniques = list(set(techniques)- set(setting_ids))
    if len(missing_technique) > 0 or len(unknown_techniques) > 0:
        invalid = True
        message = ''
        if len(missing_technique) > 0:
            message += f'Techniques are missing for experimental setting ' \
                       f'{", ".join(missing_technique)}.'
        if len(unknown_techniques) > 0:
            message += f'Techniques were given for experimental setting ' \
                       f'{", ".join(unknown_techniques)} which was not ' \
                       f'defined.'
    return invalid, message


def validate_reference_genome(organisms, reference_genome):
    """
    This function tests if the reference genome matches the organism.
    :param organisms: a list of all organisms in the metadata file
    :param reference_genome: the reference genome that was specified
    :return:
    invalid: boolean to state if the reference genome is invalid
    message: a string explaining the logical error
    """
    invalid = False
    message = None
    ref_genome_whitelist = utils.get_whitelist('reference_genome', None)
    if ref_genome_whitelist:
        if not(any(organism in ref_genome_whitelist['whitelist'] for organism in organisms)) or not any([reference_genome in ref_genome_whitelist['whitelist'][organism] for
                    organism in organisms if organism in ref_genome_whitelist['whitelist']]):
            invalid = True
            organisms = [f'\'{organism}\'' for organism in organisms]
            message = (f'The reference genome \'{reference_genome}\' does not '
                       f'match the input organism ({", ".join(organisms)}).')
    return invalid, message


def validate_donor_count(pooled, donor_count):
    """
    This function tests if the donor_count matches the value stated for pooled.
    :param pooled: a boolean stating if a sample is pooled
    :param donor_count: the donor_count of a sample
    :return:
    invalid: boolean to state if the value is invalid
    message: a string explaining the logical error
    """
    invalid = False
    message = None
    if pooled in [True, 'True'] and donor_count <= 1:
        invalid = True
        message = (f'Found donor count {donor_count} for pooled sample. '
                   f'The donor count should be greater than 1.')
    elif pooled in [False, 'False'] and donor_count > 1:
        invalid = True
        message = (f'Found donor count {donor_count} for sample that is not '
                   f'pooled. The donor count should be 1.')
    return invalid, message



import src.find_metafiles as find_metafiles
import src.utils as utils
import src.web_interface.html_output as html_output
import src.web_interface.wi_utils as wi_utils

# TODO: refactor and comment


def get_meta_info(html_str, metafiles, project_id, validation_reports):
    """
    This file creates an HTML summary for a project containing metadata
    :param project_id: the id of the project
    :return: html_str: the summary in HTML
    """
    correct_file = None
    for metafile in metafiles:
        if (
                "id" in metafile
                and metafile["id"] == project_id
        ):
            correct_file = utils.read_in_yaml(metafile['path'])
            break

    if correct_file is not None:
        html_str += f'<h2 style="text-align:center;">{project_id}</h2><hr>'
        if validation_reports is not None:
            if (
                    validation_reports["error_count"] > 0
                    or validation_reports["warning_count"] > 0
            ):
                error = None
                warning = None

                for report in validation_reports["corrupt_files"]["report"]:
                    if report["file"]["path"] == correct_file["path"]:
                        error = report["error"]
                        warning = report["warning"]
                        break

                # TODO: error Handling + Ausgabe
                if error is not None:
                    html_str += f'<font color="red"><h3><b>ERROR:</b></h3>'
                    if len(error[0]) > 0:
                        html_str += f"<b>Missing mandatory keys:</b><br>"
                        html_str += "<ul>"
                        for elem in error[0]:
                            html_str += f"<li>{elem}</li>"
                        html_str += "</ul>"
                    if len(error[1]) > 0:
                        print(error[1])
                        html_str += f"<b>Invalid keys:</b><br>"
                        html_str += "<ul>"
                        for elem in error[1]:
                            value = correct_file
                            for key in elem.split(":"):
                                if isinstance(value, list):
                                    for l_elem in value:
                                        if key in l_elem:
                                            value = l_elem[key]
                                            break
                                else:
                                    if key in value:
                                        value = value[key]
                            html_str += f"<li>{elem}: {value}</li>"
                            correct_file = wi_utils.pop_key(
                                correct_file, elem.split(":"), value
                            )
                        html_str += "</ul>"

                    if len(error[2]) > 0:
                        html_str += f"<b>Invalid entries:</b><br>"
                        html_str += "<ul>"
                        for elem in error[2]:
                            html_str += (
                                f'<li>{elem.split(":")[-1]} in '
                                f'{":".join(elem.split(":")[:-1])}</li>'
                            )
                            correct_file = wi_utils.pop_value(
                                correct_file, elem.split(":")[:-1],
                                elem.split(":")[-1]
                            )
                        html_str += "</ul>"

                    if len(error[3]) > 0:
                        html_str += f"<b>Invalid values:</b><br>"
                        html_str += "<ul>"
                        for elem in error[3]:
                            html_str += (
                                f"<li>{elem[0]}: {elem[1]} -> " f"{elem[2]}</li>"
                            )
                        html_str += "</ul>"

                    html_str += '</font><hr style="border-top: dotted 1px; background-color: transparent;" />'

                if warning is not None:
                    html_str += f'<font color="orange"><h3><b>WARNING:</b></h3>'
                    html_str += "<ul>"
                    for elem in warning:
                        message = elem[0].replace("'", "")
                        html_str += f"<li>{message}: {elem[1]}</li>"
                    html_str += "</ul>"
                    html_str += '</font><hr style="border-top: dotted 1px; background-color: transparent;" />'

        if "path" in correct_file:
            correct_file.pop("path")
        for elem in correct_file:
            end = f'{"<hr><br>" if elem != list(correct_file.keys())[-1] else ""}'
            html_str = (
                f"{html_str}<h3>{elem}</h3>"
                f"{html_output.object_to_html(correct_file[elem], 0, False)}<br>"
                f"{end}"
            )

    return html_str, correct_file


def get_search_mask(key_yaml):
    """
    This functions returns all necessary information for the search mask.
    :return: a dictionary containing all keys of the metadata structure and a
             whitelist object
    """
    keys = [
        {
            "key_name": "All keys",
            "display_name": "All Fields",
            "nested": [],
            "whitelist": False,
            "chained_keys": "",
        }
    ]
    keys += get_search_keys(key_yaml, "")
    return {"keys": keys}


def find_metadata(key_yaml, path, search_string):
    """
    This function searches for metadata files that match a search string in a
    given directory
    :param path: the path that should be searched
    :param search_string: the search string
    :return: new_files: a list containing all matching files
    """
    files = find_metafiles.find_projects(key_yaml, path, search_string, True)
    new_files = []
    for i in range(len(files)):
        for key in files[i]:
            res = {"id": key, "path": files[i][key]["path"]}
            try:
                res["project_name"] = files[i][key]["project"]["project_name"]
            except KeyError:
                res["project_name"] = None

            try:
                res["owner"] = files[i][key]["project"]["owner"]["name"]
            except KeyError:
                res["owner"] = None

            try:
                res["email"] = files[i][key]["project"]["owner"]["email"]
            except KeyError:
                res["email"] = None

            res["organisms"] = list(utils.find_keys(files[i][key], "organism_name"))

            try:
                res["description"] = files[i][key]["project"]["description"]
            except KeyError:
                res["description"] = None

            try:
                res["date"] = files[i][key]["project"]["date"]
            except KeyError:
                res["date"] = None

            if "nerd" in files[i][key]["project"]:
                nerds = []
                for nerd in files[i][key]["project"]["nerd"]:
                    nerds.append(nerd["name"])
                res["nerd"] = nerds
            else:
                res["nerd"] = None
            
            cell_type = list(set(utils.find_keys(files[i][key], 'cell_type')))
            res['cell_type'] = cell_type

            tissue = []
            
            tissues = list(utils.find_keys(files[i][key], 'tissue'))
            for elem in tissues:
                tissue += elem
            tissue = list(set(tissue))
            res['tissue'] = tissue

            # treatment
            treatment = []
            
            medical = list(utils.find_list_key(files[i][key], 'medical_treatment:treatment_type'))
            treatment += list(set(medical))
            
            physical = list(utils.find_keys(files[i][key], 'physical_treatment'))
            treatment += list(set(physical))
            
            injury = list(utils.find_list_key(files[i][key], 'injury:injury_type'))
            treatment += list(set(injury))

            res['treatment'] = treatment

            # disease

            disease = list(utils.find_list_key(files[i][key], 'disease:disease_type'))
            res['disease'] = list(set(disease))

            new_files.append(res)

    return new_files


def get_search_keys(key_yaml, chained, is_factor= False):
    """
    This function returns all keys of the metadata structure in a nested way
    :param key_yaml: the read in keys.yaml
    :param chained: the position of the key
    :return: res: a dictionary containing all metadata keys
    """
    res = []
    for key in key_yaml:
        d = {
            "key_name": key,
            "display_name": list(utils.find_keys(key_yaml, key))[0]["display_name"],
        }

        exist = True
        if (
            isinstance(key_yaml[key]["value"], dict)
            and not set(["mandatory", "list", "desc", "display_name", "value"])
            <= set(key_yaml[key]["value"].keys())
            and not (
                "special_case" in key_yaml[key]
                and "merge" in key_yaml[key]["special_case"]
            )
        ):
            if key == 'technical_replicates':
                exist = False
            else:
                d["nested"] = get_search_keys(
                    key_yaml[key]["value"],
                    f"{chained}{key}:" if chained != "" else f"{key}:",
                    True if 'samples:' in chained else False)
        else:
            if 'special_case' in key_yaml[key] and 'generated' in \
                    key_yaml[key]['special_case'] and \
                    key_yaml[key]['special_case']['generated'] in \
                    ['now', 'end'] and key != 'id' and not \
                    ('invisible' in key_yaml[key]['special_case'] and
                     key_yaml[key]['special_case']['invisible']):
                exist = False
            elif chained.endswith('experimental_factors:') and key == 'values':
                exist = False
            else:
                d["chained_keys"] = f"{chained}{key}:" if chained != "" else f"{key}:"
                d["nested"] = []

        if exist:
            if len(d['nested']) == 1:
                if key == 'experimental_factors':
                    display_name = d['display_name']
                    d = d['nested'][0]
                    d['display_name'] = display_name
                else:
                    d = d['nested'][0]
            else:
                if "whitelist" in key_yaml[key]:
                    d["whitelist"] = key_yaml[key]["whitelist"]
                elif (
                    "special_case" in key_yaml[key] and "merge" in key_yaml[key]["special_case"]
                ):
                    d["whitelist"] = key_yaml[key]["value"][
                        key_yaml[key]["special_case"]["merge"]
                    ]["whitelist"]

                if "whitelist" in d and d["whitelist"]:
                    d["search_info"] = {"key_name": key}
            if is_factor:
                display_name = chained.rstrip(':').split(':')[-1].replace('_', ' ').title()
                d['display_name'] = f'{display_name} {d["display_name"]}'
            res.append(d)
    return res



import pytz
from dateutil import parser


def pop_key(metafile, key_list, value):
    """
    This function iterates over a list of keys in order to remove the last key
    and every key above it if its remaining value is an empty list or
    dictionary
    :param metafile: the read in metadata file
    :param key_list: a list of keys
    :param value: the value within the last key
    :return: metafile: the read in metadata file without the popped key
    """

    # metafile is a list
    if isinstance(metafile, list):

        # iterate over list
        for i in range(len(metafile)):

            # call this function on every list element
            metafile[i] = pop_key(metafile[i], key_list, value)

            # if the remaining list element is an empty dictionary or list then
            # set it to None
            if (isinstance(metafile[i], dict) or
                    isinstance(metafile[i], list)) and len(metafile[i]) == 0:
                metafile[i] = None

        # remove all list elements that are None
        metafile = [x for x in metafile if x is not None]

    # metafile is a dictionary
    elif isinstance(metafile, dict):

        # only one list key is left -> must be removed
        if len(key_list) == 1:

            # test if the key is in the metafile and its value matches the key
            # to be removed
            if key_list[0] in metafile and metafile[key_list[0]] == value:

                # remove the key from the metafile
                metafile.pop(key_list[0])

        # more keys are left in the list
        else:

            # call this function on the part of the metafile within the first
            # key of the list
            metafile[key_list[0]] = pop_key(metafile[key_list[0]],
                                            key_list[1:], value)
            # test if the remaining part of the metafile is an empty dictionary
            # or list
            if (isinstance(metafile[key_list[0]], dict) or
                    isinstance(metafile[key_list[0]], list)) and \
                    len(metafile[key_list[0]]) == 0:

                # remove the key from the metafile
                metafile.pop(key_list[0])

    return metafile


def pop_value(metafile, key_list, value):
    """
    This function removes a value from the metadata object. If the removal
    leads to an empty key or list then those are removed as well
    :param metafile: the read in metadata file
    :param key_list: a list of keys
    :param value: the value to be removed
    :return: metafile: the read in metadata file without the removed value
    """

    # only one key is in the list
    if len(key_list) == 1:

        # metafile is a list
        if isinstance(metafile, list):

            # iterate over the elements of the list
            for i in range(len(metafile)):

                # call this function on every list element
                metafile[i] = pop_value(metafile[i], key_list, value)

                # test if the remaining element is an empty dictionary or list
                if (isinstance(metafile[i], dict) or
                        isinstance(metafile[i], list)) and \
                        len(metafile[i]) == 0:

                    # set the element to None
                    metafile[i] = None

            # remove all list elements that are None
            metafile = [x for x in metafile if x is not None]

        # metafile is a dictionary and contains the key
        elif key_list[0] in metafile:

            # the value of the key is a list
            if isinstance(metafile[key_list[0]], list):

                # remove the faulty value from the list
                metafile[key_list[0]] = [x for x in metafile[key_list[0]] if
                                         x != value]

                # test if the remaining list is empty then remove the key
                if len(metafile[key_list[0]]) == 0:
                    metafile.pop(key_list[0])

            # remove the key if it contains the faulty value
            elif metafile[key_list[0]] == value:
                metafile.pop(key_list[0])

    # multiple keys in the list
    else:

        # metafile is a list
        if isinstance(metafile, list):

            # iterate over the list elements
            for i in range(len(metafile)):

                # call this function on every list element
                metafile[i] = pop_value(metafile[i], key_list, value)

                # set the element to None if it contains an empty list or
                # dictionary
                if len(metafile[i]) == 0:
                    metafile[i] = None

            # remove all elements from the list that are None
            metafile = [x for x in metafile if x is not None]

        # metafile is a dictionary
        elif isinstance(metafile, dict) and key_list[0] in metafile:

            # call this function on the part of the metafile within the first
            # key of the list
            metafile[key_list[0]] = pop_value(metafile[key_list[0]],
                                              key_list[1:], value)
            # test if the remaining part of the metafile is an empty dictionary
            # or list
            if (isinstance(metafile[key_list[0]], dict) or
                    isinstance(metafile[key_list[0]], list)) and \
                    len(metafile[key_list[0]]) == 0:

                # remove the key from the metafile
                metafile.pop(key_list[0])

    return metafile


def date_to_str(date):
    """
    This function converts the date from default time in ISO 8601 format to
    time zone Berlin and changes the format to 'DD.MM.YYYY'
    :param date: the date to be converted
    :return: the converted date
    """

    # read in the date as default time
    default_time = parser.parse(date)

    # initialize the time zone for Berlin
    timezone = pytz.timezone("Europe/Berlin")

    # convert the default time to the timezone Berlin
    local_time = default_time.astimezone(timezone)

    # return the date in the format 'DD.MM.YYYY'
    return local_time.strftime("%d.%m.%Y")


def str_to_date(value):
    """
    This function converts a string containing the date in the format
    'DD.MM.YYYY' to ISO 8601 format and changes the timezone from Berlin to
    default time
    :param value: the string value containing the date
    :return: the date in ISO 8601 format
    """

    # read in the string as local time
    local_time = parser.parse(value, dayfirst=True)

    # convert the local time to default time
    default_time = local_time.astimezone(pytz.utc)

    # return the date in the format ISO 8601
    return default_time.strftime("%Y-%m-%dT%X.%fZ")


def parse_headers(headers, value, mode='dict'):
    """
    This function splits a value into a dictionary depending on the header
    :param mode: a string defining the type the split value should be returned
                 as, default is 'dict'
                 -> dict: returns the split value in a dictionary
                 -> str: returns the split value as a string
                         ('<key1>:<value1>|<keys2>:<value2>')
    :param headers: a string containing the header keys divided by space
    :param value: a string value to be split at space
    :return: new_val: the dictionary containing header keys and their values
    """

    # define a dictionary or string to save the new value to
    # (depending on mode)
    new_val = {} if mode == 'dict' else ''
    # iterate over the keys in the header
    if len(value.split(' ')) == len(headers.split(' ')):
        for key_index in range(len(headers.split(' '))):

            # return a dictionary
            if mode == 'dict':

                # save the header and value at index 'key_index' in the dictionary
                # (header and value split at ' ' -> lists that are indexed)
                new_val[headers.split(' ')[key_index]] = \
                    value.split(' ')[key_index]

            # return a string
            elif mode == 'str':

                # add the header and value at index 'key_index' to the string
                new_val = f'{new_val}{"|" if key_index > 0 else ""}' \
                        f'{headers.split(" ")[key_index]}:"' \
                        f'{value.split(" ")[key_index]}"'
    else:
        new_val = value

    return new_val


def parse_whitelist_keys(whitelist_keys, value, headers, mode='dict'):
    """
    This function removes the group-key from the end of the value of a plain
    grouped whitelist and splits the value into a dictionary depending on a
    given header
    :param mode: a string defining the type the value split according to the
                 headers should be returned as, default is 'dict'
                 -> dict: returns the split value in a dictionary
                 -> str: returns the split value as a string
                         ('<key1>:<value1>|<keys2>:<value2>')
    :param whitelist_keys: a list of keys the whitelist was grouped by
    :param value: the value that should be converted
    :param headers: a string of keys the value should be split into
                    (might be None if no header is specified)
    :return: value: the converted value (dictionary or string depending on
                    weather a header was given)
    """

    # iterate over whitelist keys
    for k in whitelist_keys:

        # remove the '(<whitelist_key>)' from the end of the value
        if value.endswith(f' ({k})'):
            value = value.replace(f' ({k})', '')

            # test if wi object contains headers
            if headers is not None and k in headers:

                # replace the original value with the one split according to
                # the header
                value = parse_headers(headers[k], value, mode=mode)

            # break since the whitelist key was found in the header
            # -> all other whitelist keys cannot be there too
            # -> better performance
            break

    return value
